/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _ASM_X86_CURRENT_H
#define _ASM_X86_CURRENT_H

#include <linux/compiler.h>
#include <asm/percpu.h>

#ifndef __ASSEMBLY__
struct task_struct;

DECLARE_PER_CPU(struct task_struct *, current_task);

static __always_inline struct task_struct *get_current(void)
{
	return this_cpu_read_stable(current_task);
}

#define current get_current()

#ifdef CONFIG_DEBUG_POISON_CURRENT_TASK
#define decode_tsk_ptr(tsk)	((struct task_struct *)((unsigned long)tsk << 1))
#define encode_tsk_ptr(tsk)	((struct task_struct *)((unsigned long)tsk >> 1))
#else
#define decode_tsk_ptr(tsk)	(tsk)
#define encode_tsk_ptr(tsk)	(tsk)
#endif

#else /* !__ASSEMBLY__ */

#define GET_CURRENT(reg) \
	_ASM_MOV PER_CPU_VAR(current_task),reg

#ifdef CONFIG_DEBUG_POISON_CURRENT_TASK
.macro DECODE_TSK_PTR tsk:req
	shlq	$1, \tsk
.endm
.macro ENCODE_TSK_PTR tsk:req
	shrq	$1, \tsk
.endm
#endif

#endif /* __ASSEMBLY__ */

#define X86_GPF_POISONED_CURRENT_TASK	_BITUL(31)

#endif /* _ASM_X86_CURRENT_H */
