/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (C) 2004  PaX Team <pageexec@freemail.hu> */

#ifndef _ASM_X86_KERNEXEC_H
#define _ASM_X86_KERNEXEC_H

#ifndef __ASSEMBLY__
#include <linux/compiler.h>
#include <linux/preempt.h>
#include <asm/bug.h>
#include <asm/special_insns.h>
#include <asm/pgtable_types.h>

#ifdef CONFIG_PAX_KERNEXEC
#ifdef CONFIG_X86_32
static inline unsigned long __turn_off_size_overflow ktla_ktva(unsigned long addr)
{
	BUG_ON(addr >= PAGE_OFFSET);
	return addr + LOAD_PHYSICAL_ADDR + PAGE_OFFSET;
}
#define ktla_ktva	ktla_ktva

static inline unsigned long __turn_off_size_overflow ktva_ktla(unsigned long addr)
{
	BUG_ON(addr < PAGE_OFFSET);
	return addr - LOAD_PHYSICAL_ADDR - PAGE_OFFSET;
}
#define ktva_ktla	ktva_ktla
#else /* !CONFIG_X86_32 */
#define ktla_ktva(addr)		(addr)
#define ktva_ktla(addr)		(addr)
#endif /* CONFIG_X86_32 */

unsigned long pax_open_kernel(void) __opens(kernel) __maybe_used __leaf_only __preserve_all_registers;
#define __HAVE_ARCH_PAX_OPEN_KERNEL

unsigned long pax_close_kernel(void) __closes(kernel) __maybe_used __leaf_only __preserve_all_registers;
#define __HAVE_ARCH_PAX_CLOSE_KERNEL
#endif /* CONFIG_PAX_KERNEXEC */

#ifdef CONFIG_PAX_KERNSEAL
void *pax_expose_page(struct page *page) __maybe_used;
void *pax_expose_page_user(struct page *page) __maybe_used;
void *pax_expose_page_nested(struct page *page) __maybe_used;
void *pax_expose_page_nested_user(struct page *page) __maybe_used;
void *pax_expose_page_prot(struct page *page, pgprot_t prot) __maybe_used;
#define __HAVE_ARCH_PAX_EXPOSE_KERNEL

void pax_hide_page(const void *address) __maybe_used;
void pax_hide_page_nested(const void *address) __maybe_used;
#define __HAVE_ARCH_PAX_HIDE_KERNEL
#endif /* CONFIG_PAX_KERNSEAL */

#include <asm-generic/kernexec.h>

#endif /* __ASSEMBLY__ */

#endif /* _ASM_X86_KERNEXEC_H */
