/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _ASM_X86_PGTABLE_32_TYPES_H
#define _ASM_X86_PGTABLE_32_TYPES_H

/*
 * The Linux x86 paging architecture is 'compile-time dual-mode', it
 * implements both the traditional 2-level x86 page tables and the
 * newer 3-level PAE-mode page tables.
 */
#ifdef CONFIG_X86_PAE
# include <asm/pgtable-3level_types.h>
# define PMD_SIZE	(_AC(1, UL) << PMD_SHIFT)
# define PMD_MASK	(~(PMD_SIZE - 1))
#else
# include <asm/pgtable-2level_types.h>
#endif

#define pgtable_l5_enabled() 0

#define PGDIR_SIZE	(1UL << PGDIR_SHIFT)
#define PGDIR_MASK	(~(PGDIR_SIZE - 1))

#ifdef CONFIG_PAX_KERNEXEC
#ifndef __ASSEMBLY__
extern unsigned char MODULES_EXEC_VADDR[];
extern unsigned char MODULES_EXEC_END[];

extern unsigned char __LOAD_PHYSICAL_ADDR[];
#define LOAD_PHYSICAL_ADDR ((unsigned long)__LOAD_PHYSICAL_ADDR)
#endif /* __ASSEMBLY__ */
#else /* !CONFIG_PAX_KERNEXEC */
#define MODULES_EXEC_VADDR	VMALLOC_START
#define MODULES_EXEC_END	VMALLOC_END
#endif

#endif /* _ASM_X86_PGTABLE_32_TYPES_H */
