/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (C) 2004  PaX Team <pageexec@freemail.hu> */

#ifndef _ASM_GENERIC_KERNEXEC_H
#define _ASM_GENERIC_KERNEXEC_H

#ifndef __ASSEMBLY__

#ifdef CONFIG_PAX_KERNEXEC
/* sanity checks under KERNEXEC */
#ifndef ktla_ktva
#error KERNEXEC requires ktla_ktva
#endif

#ifndef ktva_ktla
#error KERNEXEC requires ktva_ktla
#endif

#ifndef __HAVE_ARCH_PAX_OPEN_KERNEL
#error KERNEXEC requires pax_open_kernel
#endif

#ifndef __HAVE_ARCH_PAX_CLOSE_KERNEL
#error KERNEXEC requires pax_close_kernel
#endif
#else /* !CONFIG_PAX_KERNEXEC */
#define ktla_ktva(addr)		(addr)
#define ktva_ktla(addr)		(addr)

static __always_inline __opens(kernel) unsigned long pax_open_kernel(void)
{
	__open(kernel);
	return 0;
}

static __always_inline __closes(kernel) unsigned long pax_close_kernel(void)
{
	__close(kernel);
	return 0;
}
#endif /* CONFIG_PAX_KERNEXEC */

#ifdef CONFIG_PAX_KERNSEAL
#ifndef __HAVE_ARCH_PAX_EXPOSE_KERNEL
#error KERNSEAL requires pax_expose_kernel
#endif

#ifndef __HAVE_ARCH_PAX_EXPOSE_KERNEL
#error KERNSEAL requires pax_expose_kernel
#endif

static __always_inline __opens(kernel) unsigned long pax_open_seal(void)
{
	return pax_open_kernel();
}

static __always_inline __closes(kernel) unsigned long pax_close_seal(void)
{
	return pax_close_kernel();
}
#else /* !CONFIG_PAX_KERNSEAL */
static __always_inline __opens(kernel) unsigned long pax_open_seal(void)
{
	__open(kernel);
	return 0;
}
static __always_inline __closes(kernel) unsigned long pax_close_seal(void)
{
	__close(kernel);
	return 0;
}
#endif /* CONFIG_PAX_KERNSEAL */

#if defined(CONFIG_PAX_KERNEXEC) && !defined(CONFIG_PAX_KERNSEAL)
static __always_inline __opens(kernel) unsigned long pax_open_kernexec(void)
{
	return pax_open_kernel();
}

static __always_inline __closes(kernel) unsigned long pax_close_kernexec(void)
{
	return pax_close_kernel();
}
#else /* !CONFIG_PAX_KERNEXEC || CONFIG_PAX_KERNSEAL */
static __always_inline __opens(kernel) unsigned long pax_open_kernexec(void)
{
	__open(kernel);
	return 0;
}

static __always_inline __closes(kernel) unsigned long pax_close_kernexec(void)
{
	__close(kernel);
	return 0;
}
#endif /* CONFIG_PAX_KERNEXEC && !CONFIG_PAX_KERNSEAL */

#endif /* __ASSEMBLY__ */

#endif /* _ASM_GENERIC_KERNEXEC_H */
