/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _LINUX_ONCE_LITE_H
#define _LINUX_ONCE_LITE_H

#ifndef __ASSEMBLY__
#include <linux/types.h>

bool noinstr __do_once_lite(atomic_t *done);

/* Call a function once. Similar to DO_ONCE(), but does not use jump label
 * patching via static keys.
 */
#define DO_ONCE_LITE(func, ...)						\
	DO_ONCE_LITE_IF(true, func, ##__VA_ARGS__)

#define __ONCE_LITE_IF(condition)					\
	({								\
		static atomic_t __section(".data.once") __already_done = ATOMIC_INIT(0);\
		unlikely(!!(condition) && __do_once_lite(&__already_done));\
	})

#define DO_ONCE_LITE_IF(condition, func, ...)				\
	({								\
		bool __ret_do_once = !!(condition);			\
									\
		if (__ONCE_LITE_IF(__ret_do_once))			\
			func(__VA_ARGS__);				\
									\
		unlikely(__ret_do_once);				\
	})
#endif /* __ASSEMBLY__ */

#endif /* _LINUX_ONCE_LITE_H */
