/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _LINUX_STACKLEAK_H
#define _LINUX_STACKLEAK_H

#include <linux/sched.h>
#include <linux/sched/task_stack.h>

/*
 * Check that the poison value points to the unused hole in the
 * virtual memory map for your platform.
 */
#define STACKLEAK_POISON -0xBEEF
#define STACKLEAK_SEARCH_DEPTH (2 * SMP_CACHE_BYTES)
#ifdef CONFIG_PPC64
#define STACKLEAK_LOWEST_STACK (STACKLEAK_SEARCH_DEPTH + STACK_FRAME_WITH_PT_REGS + KERNEL_REDZONE_SIZE)
#else
#define STACKLEAK_LOWEST_STACK (STACKLEAK_SEARCH_DEPTH + sizeof(struct pt_regs))
#endif

#ifdef CONFIG_GCC_PLUGIN_STACKLEAK
#include <asm/stacktrace.h>

static inline void stackleak_task_init(struct task_struct *t)
{
	t->lowest_stack = (unsigned long)end_of_stack(t) + sizeof(unsigned long);
# ifdef CONFIG_STACKLEAK_METRICS
	t->prev_lowest_stack = t->lowest_stack;
# endif
}

#ifdef CONFIG_STACKLEAK_RUNTIME_DISABLE
int stack_erasing_sysctl(struct ctl_table *table, int write,
			void *buffer, size_t *lenp, loff_t *ppos);
#endif

asmlinkage void noinstr __nostackprotector stackleak_erase(void);

#else /* !CONFIG_GCC_PLUGIN_STACKLEAK */
static inline void stackleak_task_init(struct task_struct *t) { }
#endif

#endif
