/* SPDX-License-Identifier: GPL-2.0 */
#ifndef GCC_COMMON_H_INCLUDED
#define GCC_COMMON_H_INCLUDED

#include "bversion.h"

#if BUILDING_GCC_VERSION >= 8000
#define IN_TARGET_CODE
#endif

#if BUILDING_GCC_VERSION >= 6000
#include "gcc-plugin.h"
#else
#include "plugin.h"
#endif
#include "plugin-version.h"

#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "line-map.h"
#include "input.h"
#if BUILDING_GCC_VERSION <= 4008
#define thread_local _thread_local
#endif
#include "tree.h"
#include "vec.h"
#if BUILDING_GCC_VERSION <= 4008
#undef thread_local
#endif

#include "tree-inline.h"
#include "version.h"
#include "rtl.h"

#if BUILDING_GCC_VERSION >= 7000
#include "memmodel.h"
#endif

#include "tm_p.h"
#include "flags.h"
//#include "insn-attr.h"
//#include "insn-codes.h"
//#include "insn-config.h"
//#include "insn-flags.h"
#include "hard-reg-set.h"
//#include "recog.h"
#include "output.h"
#include "except.h"
#include "function.h"
#include "toplev.h"

#if BUILDING_GCC_VERSION >= 5000
#include "expr.h"
#include "explow.h"
#else
rtx copy_to_reg(rtx);
void expand_builtin_trap(void);
tree array_ref_element_size(tree);
tree component_ref_field_offset(tree);
#endif

#include "basic-block.h"
#include "intl.h"
#include "ggc.h"

#if BUILDING_GCC_VERSION >= 5000
#include "graph.h"
#include "regs.h"
#endif

#include "timevar.h"
#if BUILDING_GCC_VERSION < 10000
#include "params.h"
#endif

#if BUILDING_GCC_VERSION <= 4009
#include "pointer-set.h"
#else
#include "hash-map.h"
#endif

#include "emit-rtl.h"
//#include "reload.h"
//#include "ira.h"
//#include "dwarf2asm.h"
#include "debug.h"
#include "target.h"
#include "langhooks.h"
#include "cfgloop.h"

#if BUILDING_GCC_VERSION >= 5000
#include "langhooks-def.h"
#include "targhooks.h"
#include "tree-scalar-evolution.h"
#else
void scev_initialize(void);
void scev_finalize(void);
enum machine_mode default_addr_space_pointer_mode (addr_space_t);
enum machine_mode default_addr_space_address_mode (addr_space_t);
bool default_addr_space_valid_pointer_mode(enum machine_mode mode, addr_space_t as);
bool default_addr_space_legitimate_address_p(enum machine_mode mode, rtx mem, bool strict, addr_space_t as);
rtx default_addr_space_legitimize_address(rtx x, rtx oldx, enum machine_mode mode, addr_space_t as);
tree default_stack_protect_guard(void);
#if BUILDING_GCC_VERSION >= 4007
rtx default_function_arg(cumulative_args_t, enum machine_mode, const_tree, bool);
#else
rtx default_function_arg(CUMULATIVE_ARGS *, enum machine_mode, const_tree, bool);
#endif
#endif

#if BUILDING_GCC_VERSION >= 4009
#include "tree-ssa-loop-niter.h"
#include "gimple-builder.h"
#include "calls.h"
#endif

//#include "hosthooks.h"
#include "cgraph.h"
#include "opts.h"
//#include "coverage.h"
//#include "value-prof.h"

#if BUILDING_GCC_VERSION == 4005
#include <sys/mman.h>
#endif

#if BUILDING_GCC_VERSION >= 4007
#include "tree-pretty-print.h"
#include "gimple-pretty-print.h"
#endif

#if BUILDING_GCC_VERSION >= 4006
/*
 * The c-family headers were moved into a subdirectory in GCC version
 * 4.7, but most plugin-building users of GCC 4.6 are using the Debian
 * or Ubuntu package, which has an out-of-tree patch to move this to the
 * same location as found in 4.7 and later:
 * https://sources.debian.net/src/gcc-4.6/4.6.3-14/debian/patches/pr45078.diff/
 */
//#include "c-tree.h"
//#include "cp/cp-tree.h"
#include "c-family/c-common.h"
#include "c-family/c-pragma.h"
#else
#include "c-common.h"
#include "c-pragma.h"
#endif

#if BUILDING_GCC_VERSION <= 4008
#include "tree-flow.h"
#else
#include "tree-cfgcleanup.h"
#include "tree-ssa-operands.h"
#include "tree-into-ssa.h"
#include "tree-dfa.h"
// partition.h is not installed...
//#include "tree-ssa-live.h"
#endif

#if BUILDING_GCC_VERSION >= 4008
#include "is-a.h"
#endif

#include "diagnostic.h"
//#include "tree-diagnostic.h"
#include "tree-dump.h"
#include "tree-pass.h"

#if BUILDING_GCC_VERSION >= 4009
#include "pass_manager.h"
#include "df.h"
#endif

#include "predict.h"

#if BUILDING_GCC_VERSION >= 4009
#include "stringpool.h"
#include "attribs.h"
#include "varasm.h"
#include "stor-layout.h"
#include "internal-fn.h"
#include "gimple-expr.h"
//#include "diagnostic-color.h"
#include "context.h"
#include "tree-ssa-alias.h"
#include "tree-ssa.h"

#if BUILDING_GCC_VERSION >= 7000
#include "tree-vrp.h"
#endif

#include "print-tree.h"
#include "tree-eh.h"
#include "stmt.h"
#include "gimplify.h"
#endif

#include "gimple.h"

#include "ipa-utils.h"
#if BUILDING_GCC_VERSION >= 5000
# if __GNUC__ >= 12
#  pragma GCC diagnostic push
#  pragma GCC diagnostic ignored "-Wmissing-template-keyword"
# endif
#include "symbol-summary.h"
# if __GNUC__ >= 12
#  pragma GCC diagnostic pop
# endif
#endif
#include "ipa-prop.h"
#if BUILDING_GCC_VERSION >= 8000
#include "ipa-fnsummary.h"
#endif
#if BUILDING_GCC_VERSION >= 5000
#include "ipa-inline.h"
#endif

#if BUILDING_GCC_VERSION >= 4009
#include "tree-ssanames.h"
#include "tree-ssa-loop.h"
#include "tree-ssa-loop-manip.h"
#include "tree-phinodes.h"
#include "tree-cfg.h"
#include "gimple-iterator.h"
#include "gimple-fold.h"
#include "gimple-ssa.h"
#include "gimple-walk.h"
#include "ssa-iterators.h"
#include "gimplify-me.h"
#endif

//#include "lto/lto.h"
#if BUILDING_GCC_VERSION >= 4007
//#include "data-streamer.h"
//#include "streamer-hooks.h"
#else
//#include "lto-streamer.h"
#endif
//#include "lto-compress.h"
#if BUILDING_GCC_VERSION >= 5000
//#include "lto-section-names.h"
#include "builtins.h"
#include "domwalk.h"
#endif

#if BUILDING_GCC_VERSION >= 6000
typedef gimple *gimple_stmt_ptr;
typedef const gimple *const_gimple_stmt_ptr;
#else
typedef gimple gimple_stmt_ptr;
typedef const_gimple const_gimple_stmt_ptr;
#endif

#if BUILDING_GCC_VERSION >= 5000
#include "tree-ssa-propagate.h"
#else
bool may_propagate_copy_into_stmt(gimple_stmt_ptr, tree);
#endif

#if BUILDING_GCC_VERSION >= 5000
#include "asan.h"
#elif BUILDING_GCC_VERSION >= 4008
void initialize_sanitizer_builtins(void);
#endif

#if BUILDING_GCC_VERSION >= 8000
#include "profile-count.h"
#include "poly-int.h"
#include "tree-ssa-sccvn.h"
#endif

#if BUILDING_GCC_VERSION >= 11000
#include "symtab-clones.h"
#include "symtab-thunks.h"
#include "gimple-range.h"
#include "vr-values.h"
#endif

#if BUILDING_GCC_VERSION >= 12000
#include "gimple-range-path.h"
#endif

#if BUILDING_GCC_VERSION >= 5000
#include "dbgcnt.h"
#else
#endif

/* missing from basic_block.h... */
void debug_dominance_info(enum cdi_direction dir);
void debug_dominance_tree(enum cdi_direction dir, basic_block root);

/* since tree-ssa-live.h is not usable due to partition.h not being installed... */
void dump_scope_blocks(FILE *file, int flags);
void debug_scope_block(tree scope, int flags);
void debug_scope_blocks(int flags);

#if BUILDING_GCC_VERSION == 4006
void debug_gimple_stmt(gimple_stmt_ptr);
void debug_gimple_seq(gimple_seq);
void print_gimple_seq(FILE *, gimple_seq, int, int);
void print_gimple_stmt(FILE *, gimple_stmt_ptr, int, int);
void print_gimple_expr(FILE *, gimple_stmt_ptr, int, int);
void dump_gimple_stmt(pretty_printer *, gimple_stmt_ptr, int, int);
#endif

// c-tree.h
tree build_c_cast(location_t, tree, tree);

// these are in graph.h that isn't installed before gcc 5...
#if BUILDING_GCC_VERSION <= 4007
void print_rtl_graph_with_bb(const char *, rtx);
#endif

#if BUILDING_GCC_VERSION >= 4008 && BUILDING_GCC_VERSION < 5000
void print_graph_cfg(const char *, struct function *);
#endif

// STATIC_ASSERT() as of 81a12b7606c5 ("Use static_assert for STATIC_ASSERT for C++11 onwards")
#ifndef STATIC_ASSERT
/* static_assert (COND, MESSAGE) is available in C++11 onwards.  */
#if __cplusplus >= 201103L
#define STATIC_ASSERT(X) \
	static_assert ((X), #X)
#else
#define STATIC_ASSERT(X) \
	typedef int assertion1[(X) ? 1 : -1] ATTRIBUTE_UNUSED
#endif
#endif

// libcpp/include/mkdeps.h isn't installed for plugins
#if BUILDING_GCC_VERSION < 10000
void deps_add_dep(struct deps *, const char *);
#else
void deps_add_dep(class mkdeps *, const char *);
#endif

#ifndef __unused
#define __unused __attribute__((unused))
#endif

#ifndef __visible
#ifdef _WIN32
#define __visible __declspec(dllexport)
#else
#define __visible __attribute__((visibility("default")))
#endif
#endif

#define __weak __attribute__((weak)) __visible

// A value high enough to not clash with any of gcc's internal builtins
#if BUILDING_GCC_VERSION < 5000
// function_code used to be a bitfield, only 11 bits wide
#define BUILT_IN_PLUGINS_START	0x7c0	// 63 custom builtins max!
#elif BUILDING_GCC_VERSION < 10000
// extended to 12 bits in c1b7a563e061 ("AVX-512. 81.1/n. Extend `function_code' field `tree_var_decl' structure.")
#define BUILT_IN_PLUGINS_START	0xfc0	// 63 custom builtins max!
#else
// finally 32 bit since 55f863c4d694 ("Make function_code a 32-bit field")
#define BUILT_IN_PLUGINS_START	0xc0ffee
#endif
#define BUILT_IN_CANARY_VALUE	(BUILT_IN_PLUGINS_START + 0)
#define BUILT_IN_CANARY_COPY	(BUILT_IN_PLUGINS_START + 1)

#define DECL_NAME_POINTER(node) IDENTIFIER_POINTER(DECL_NAME(node))
#define DECL_NAME_LENGTH(node) IDENTIFIER_LENGTH(DECL_NAME(node))
#define TYPE_NAME_POINTER(node) IDENTIFIER_POINTER(type_name(node))
#define TYPE_NAME_LENGTH(node) IDENTIFIER_LENGTH(type_name(node))
#define FUNC_OR_METHOD_POINTER_TYPE_P(TYPE) (POINTER_TYPE_P(TYPE) && FUNC_OR_METHOD_TYPE_P(TREE_TYPE(TYPE)))

/* should come from c-tree.h if only it were installed for gcc 4.5... */
#define C_TYPE_FIELDS_READONLY(TYPE) TREE_LANG_FLAG_1(TYPE)

static inline tree build_const_char_string(int len, const char *str)
{
	tree cstr, elem, index, type;

	cstr = build_string(len, str);
	elem = build_type_variant(char_type_node, 1, 0);
	index = build_index_type(size_int(len - 1));
	type = build_array_type(elem, index);
	build_pointer_type(type);
	TREE_TYPE(cstr) = type;
	TREE_CONSTANT(cstr) = 1;
	TREE_READONLY(cstr) = 1;
	TREE_STATIC(cstr) = 1;
	return cstr;
}

static inline bool plugin_version_check(struct plugin_gcc_version *v1, struct plugin_gcc_version *v2)
{
#ifdef RELAXED_VERSION_CHECK
	return !strcmp(v1->basever, v2->basever);
#else
	return plugin_default_version_check(v1, v2);
#endif
}

#define plugin_default_version_check plugin_version_check

static inline void error_gcc_version(struct plugin_gcc_version *version)
{
#ifdef RELAXED_VERSION_CHECK
	if (strcmp(gcc_version.basever, version->basever))
		error(G_("incompatible gcc version: plugin needs %s but has %s"),
		      gcc_version.basever, version->basever);
#else
	if (strcmp(gcc_version.basever, version->basever) ||
	    strcmp(gcc_version.datestamp, version->datestamp) ||
	    strcmp(gcc_version.devphase, version->devphase) ||
	    strcmp(gcc_version.revision, version->revision))
		error(G_("incompatible gcc version: plugin needs %s %s %s %s but has %s %s %s %s.\n"
			 "If this was intentional or expected, consider enabling "
			 "CONFIG_GCC_RELAXED_VERSION_CHECK after reviewing its kernel config help."),
		      gcc_version.basever, gcc_version.datestamp, gcc_version.devphase, gcc_version.revision,
		      version->basever, version->datestamp, version->devphase, version->revision);
	else
		error(G_("incompatible gcc version: plugin needs a configure string of:\n%s\nbut has:\n%s\n"
			 "If this was intentional or expected, consider enabling "
			 "CONFIG_GCC_RELAXED_VERSION_CHECK after reviewing its kernel config help."),
		      gcc_version.configuration_arguments, version->configuration_arguments);
#endif
}

static inline void __add_type_attr(tree type, const char *attr, tree args)
{
	tree oldattr;

	if (type == NULL_TREE)
		return;
	oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
	if (oldattr != NULL_TREE) {
		gcc_assert(TREE_VALUE(oldattr) == args || TREE_VALUE(TREE_VALUE(oldattr)) == TREE_VALUE(args));
		return;
	}

	TYPE_ATTRIBUTES(type) = copy_list(TYPE_ATTRIBUTES(type));
	TYPE_ATTRIBUTES(type) = tree_cons(get_identifier(attr), args, TYPE_ATTRIBUTES(type));
}

static inline void add_type_attr(tree type, const char *attr, tree args)
{
	tree main_variant = TYPE_MAIN_VARIANT(type);

	__add_type_attr(TYPE_CANONICAL(type), attr, args);
	__add_type_attr(TYPE_CANONICAL(main_variant), attr, args);
	__add_type_attr(main_variant, attr, args);

	for (type = TYPE_NEXT_VARIANT(main_variant); type; type = TYPE_NEXT_VARIANT(type)) {
		tree oldattr;

		oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
		if (oldattr != NULL_TREE)
			gcc_assert(TREE_VALUE(oldattr) == args || TREE_VALUE(TREE_VALUE(oldattr)) == TREE_VALUE(args));
		else
			TYPE_ATTRIBUTES(type) = TYPE_ATTRIBUTES(main_variant);

		__add_type_attr(TYPE_CANONICAL(type), attr, args);
	}
}

static inline void __update_type_attr(tree type, const char *attr, tree args)
{
	tree oldattr;

	if (type == NULL_TREE)
		return;
	oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
	if (oldattr == NULL_TREE)
		return;

	TYPE_ATTRIBUTES(type) = copy_list(TYPE_ATTRIBUTES(type));
	oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
	gcc_assert(oldattr);
	TREE_VALUE(oldattr) = args;
}

static inline void update_type_attr(tree type, const char *attr, tree args)
{
	tree main_variant = TYPE_MAIN_VARIANT(type);

	__update_type_attr(TYPE_CANONICAL(type), attr, args);
	__update_type_attr(TYPE_CANONICAL(main_variant), attr, args);
	__update_type_attr(main_variant, attr, args);

	for (type = TYPE_NEXT_VARIANT(main_variant); type; type = TYPE_NEXT_VARIANT(type)) {
		tree oldattr;

		oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
		if (oldattr != NULL_TREE)
			gcc_assert(TREE_VALUE(oldattr) == args || TREE_VALUE(TREE_VALUE(oldattr)) == TREE_VALUE(args));
		else
			TYPE_ATTRIBUTES(type) = TYPE_ATTRIBUTES(main_variant);

		__add_type_attr(TYPE_CANONICAL(type), attr, args);
	}
}

static inline bool plugin_register_event(const char *plugin_name, const char *event_name, plugin_callback_func callback, void *user_data)
{
	int event;

	event = get_named_event_id(event_name, NO_INSERT);
	if (event != -1)
		return false;

	event = get_named_event_id(event_name, INSERT);
	if (event < PLUGIN_EVENT_FIRST_DYNAMIC)
		return false;

	register_callback(plugin_name, event, callback, user_data);
	return true;
}

static inline bool plugin_send_event(const char *event_name, void *data)
{
	int event;

	event = get_named_event_id(event_name, NO_INSERT);
	if (event == -1)
		return false;

	invoke_plugin_callbacks(event, data);
	return true;
}

// TODO: use TYPE_IDENTIFIER instead?
static inline const_tree type_name(const_tree type)
{
	const_tree name;

	name = TYPE_NAME(type);
	if (!name)
		return NULL_TREE;

	switch (TREE_CODE(name)) {
	case IDENTIFIER_NODE:
		return name;

	case TYPE_DECL:
		gcc_assert(DECL_NAME(name));
		return DECL_NAME(name);

	default:
		gcc_unreachable();
	}
}

static inline basic_block plugin_split_edge(edge e)
{
	basic_block bb;

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));

	bb = split_edge(e);
	gcc_assert(single_succ_p(e->dest));

	if (!dom_info_available_p(CDI_POST_DOMINATORS))
		return bb;

	// split_edge doesn't adjust CDI_POST_DOMINATORS but we need this information
	set_immediate_dominator(CDI_POST_DOMINATORS, bb, single_succ(bb));

	if (get_immediate_dominator(CDI_POST_DOMINATORS, single_pred(bb)) == single_succ(bb)) {
		edge_iterator ei;
		edge f;

		FOR_EACH_EDGE(f, ei, single_pred(bb)->succs) {
			if (f == single_pred_edge(bb))
				continue;

			if (!dominated_by_p(CDI_POST_DOMINATORS, f->dest, single_pred(bb)))
				break;
		}

		if (!f)
			set_immediate_dominator(CDI_POST_DOMINATORS, single_pred(bb), bb);
	}

	return bb;
}

#if BUILDING_GCC_VERSION >= 4009
#define PASS_INFO(NAME, REF, ID, POS)		\
struct register_pass_info NAME##_pass_info = {	\
	.pass = make_pass_##NAME(g),		\
	.reference_pass_name = REF,		\
	.ref_pass_instance_number = ID,		\
	.pos_op = POS,				\
}
#else
#define PASS_INFO(NAME, REF, ID, POS)		\
struct register_pass_info NAME##_pass_info = {	\
	.pass = make_pass_##NAME(),		\
	.reference_pass_name = REF,		\
	.ref_pass_instance_number = ID,		\
	.pos_op = POS,				\
}
#endif

#if BUILDING_GCC_VERSION < 4007
#define ATTRIBUTE_SPEC_INIT(NAME, MIN, MAX, DECL, TYPE, FUNC, HANDLER, IDENTITY, EXCLUDE) \
{							\
	.name				= NAME,		\
	.min_length			= MIN,		\
	.max_length			= MAX,		\
	.decl_required			= DECL,		\
	.type_required			= TYPE,		\
	.handler			= HANDLER,	\
}
#elif BUILDING_GCC_VERSION >= 4007 && BUILDING_GCC_VERSION < 8000
#define ATTRIBUTE_SPEC_INIT(NAME, MIN, MAX, DECL, TYPE, FUNC, HANDLER, IDENTITY, EXCLUDE) \
{ \
	.name				= NAME,		\
	.min_length			= MIN,		\
	.max_length			= MAX,		\
	.decl_required			= DECL,		\
	.type_required			= TYPE,		\
	.function_type_required		= FUNC,		\
	.handler			= HANDLER,	\
	.affects_type_identity		= IDENTITY,	\
}
#else
#define ATTRIBUTE_SPEC_INIT(NAME, MIN, MAX, DECL, TYPE, FUNC, HANDLER, IDENTITY, EXCLUDE) \
{ \
	.name				= NAME,		\
	.min_length			= MIN,		\
	.max_length			= MAX,		\
	.decl_required			= DECL,		\
	.type_required			= TYPE,		\
	.function_type_required		= FUNC,		\
	.affects_type_identity		= IDENTITY,	\
	.handler			= HANDLER,	\
	.exclude			= EXCLUDE,	\
}
#endif

#define INIT_GGC_ROOT_TAB_TREE(NAME)	\
{ \
	.base = &NAME,			\
	.nelt = 1,			\
	.stride = sizeof(NAME),		\
	.cb = &gt_ggc_mx_tree_node,	\
	.pchw = &gt_pch_nx_tree_node	\
}

#define INIT_GGC_ROOT_TAB_TREES(NAME, FIELD) \
{ \
	.base = &NAME[0].FIELD,		\
	.nelt = ARRAY_SIZE(NAME),	\
	.stride = sizeof(NAME[0]),	\
	.cb = &gt_ggc_mx_tree_node,	\
	.pchw = &gt_pch_nx_tree_node	\
}

#define INIT_GGC_ROOT_TAB_RTX(NAME)	\
{ \
	.base = &NAME,			\
	.nelt = 1,			\
	.stride = sizeof(NAME),		\
	.cb = &gt_ggc_mx_rtx_def,	\
	.pchw = &gt_pch_nx_rtx_def	\
}

#if BUILDING_GCC_VERSION == 4005
#define FOR_EACH_LOCAL_DECL(FUN, I, D)			\
	for (tree vars = (FUN)->local_decls, (I) = 0;	\
		vars && ((D) = TREE_VALUE(vars));	\
		vars = TREE_CHAIN(vars), (I)++)
#define DECL_CHAIN(NODE) (TREE_CHAIN(DECL_MINIMAL_CHECK(NODE)))
#define FOR_EACH_VEC_ELT(T, V, I, P) \
	for ((I) = 0; VEC_iterate(T, (V), (I), (P)); ++(I))
#define TODO_rebuild_cgraph_edges 0
#define SCOPE_FILE_SCOPE_P(EXP) (!(EXP))

#ifndef O_BINARY
#define O_BINARY 0
#endif

static inline bool seen_error(void)
{
	return errorcount || sorrycount;
}

typedef struct varpool_node *varpool_node_ptr;

static inline varpool_node_ptr varpool_get_node(tree decl)
{
	varpool_node_ptr node;

	for (node = varpool_nodes; node ; node = node->next)
		if (node->decl == decl)
			break;

	return node;

//	this isn't exactly equivalent because it also creates a node if necessary
//	return varpool_node(decl);
}

static inline bool gimple_call_builtin_p(gimple_stmt_ptr stmt, enum built_in_function code)
{
	tree fndecl;

	if (!is_gimple_call(stmt))
		return false;
	fndecl = gimple_call_fndecl(stmt);
	if (!fndecl || DECL_BUILT_IN_CLASS(fndecl) != BUILT_IN_NORMAL)
		return false;
//	print_node(stderr, "pax", fndecl, 4);
	return DECL_FUNCTION_CODE(fndecl) == code;
}

static inline bool is_simple_builtin(tree decl)
{
	if (!decl || DECL_BUILT_IN_CLASS(decl) != BUILT_IN_NORMAL)
		return false;

	switch (DECL_FUNCTION_CODE(decl)) {
	/* Builtins that expand to constants. */
	case BUILT_IN_CONSTANT_P:
	case BUILT_IN_EXPECT:
	case BUILT_IN_OBJECT_SIZE:
	case BUILT_IN_UNREACHABLE:
	/* Simple register moves or loads from stack. */
	case BUILT_IN_RETURN_ADDRESS:
	case BUILT_IN_EXTRACT_RETURN_ADDR:
	case BUILT_IN_FROB_RETURN_ADDR:
	case BUILT_IN_RETURN:
	case BUILT_IN_AGGREGATE_INCOMING_ADDRESS:
	case BUILT_IN_FRAME_ADDRESS:
	case BUILT_IN_VA_END:
	case BUILT_IN_STACK_SAVE:
	case BUILT_IN_STACK_RESTORE:
	/* Exception state returns or moves registers around. */
	case BUILT_IN_EH_FILTER:
	case BUILT_IN_EH_POINTER:
	case BUILT_IN_EH_COPY_VALUES:
		return true;

	default:
		return false;
	}
}

static inline bool is_inexpensive_builtin(tree decl)
{
	if (!decl)
		return false;

	if (DECL_BUILT_IN_CLASS(decl) == BUILT_IN_MD)
		return true;

	if (DECL_BUILT_IN_CLASS(decl) != BUILT_IN_NORMAL)
		return false;

	switch (DECL_FUNCTION_CODE(decl)) {
	case BUILT_IN_ABS:
	case BUILT_IN_ALLOCA:
	case BUILT_IN_BSWAP32:
	case BUILT_IN_BSWAP64:
	case BUILT_IN_CLZ:
	case BUILT_IN_CLZIMAX:
	case BUILT_IN_CLZL:
	case BUILT_IN_CLZLL:
	case BUILT_IN_CTZ:
	case BUILT_IN_CTZIMAX:
	case BUILT_IN_CTZL:
	case BUILT_IN_CTZLL:
	case BUILT_IN_FFS:
	case BUILT_IN_FFSIMAX:
	case BUILT_IN_FFSL:
	case BUILT_IN_FFSLL:
	case BUILT_IN_IMAXABS:
	case BUILT_IN_FINITE:
	case BUILT_IN_FINITEF:
	case BUILT_IN_FINITEL:
	case BUILT_IN_FINITED32:
	case BUILT_IN_FINITED64:
	case BUILT_IN_FINITED128:
	case BUILT_IN_FPCLASSIFY:
	case BUILT_IN_ISFINITE:
	case BUILT_IN_ISINF_SIGN:
	case BUILT_IN_ISINF:
	case BUILT_IN_ISINFF:
	case BUILT_IN_ISINFL:
	case BUILT_IN_ISINFD32:
	case BUILT_IN_ISINFD64:
	case BUILT_IN_ISINFD128:
	case BUILT_IN_ISNAN:
	case BUILT_IN_ISNANF:
	case BUILT_IN_ISNANL:
	case BUILT_IN_ISNAND32:
	case BUILT_IN_ISNAND64:
	case BUILT_IN_ISNAND128:
	case BUILT_IN_ISNORMAL:
	case BUILT_IN_ISGREATER:
	case BUILT_IN_ISGREATEREQUAL:
	case BUILT_IN_ISLESS:
	case BUILT_IN_ISLESSEQUAL:
	case BUILT_IN_ISLESSGREATER:
	case BUILT_IN_ISUNORDERED:
	case BUILT_IN_VA_ARG_PACK:
	case BUILT_IN_VA_ARG_PACK_LEN:
	case BUILT_IN_VA_COPY:
	case BUILT_IN_TRAP:
	case BUILT_IN_SAVEREGS:
	case BUILT_IN_POPCOUNTL:
	case BUILT_IN_POPCOUNTLL:
	case BUILT_IN_POPCOUNTIMAX:
	case BUILT_IN_POPCOUNT:
	case BUILT_IN_PARITYL:
	case BUILT_IN_PARITYLL:
	case BUILT_IN_PARITYIMAX:
	case BUILT_IN_PARITY:
	case BUILT_IN_LABS:
	case BUILT_IN_LLABS:
	case BUILT_IN_PREFETCH:
		return true;

	default:
		return is_simple_builtin(decl);
	}
}

static inline void add_local_decl(struct function *fun, tree d)
{
	gcc_assert(TREE_CODE(d) == VAR_DECL);
	fun->local_decls = tree_cons(NULL_TREE, d, fun->local_decls);
}

static inline tree build1_loc(location_t loc, enum tree_code code, tree type, tree arg1)
{
	tree t = build1(code, type, arg1);
	SET_EXPR_LOCATION(t, loc);
	return t;
}

static inline tree build2_loc(location_t loc, enum tree_code code, tree type, tree arg1, tree arg2)
{
	tree t = build2(code, type, arg1, arg2);
	SET_EXPR_LOCATION(t, loc);
	return t;
}

static inline bool check_aligned_type(const_tree cand, const_tree base, unsigned int align)
{
	if (TYPE_QUALS(cand) != TYPE_QUALS(base))
		return false;

	if (TYPE_NAME(cand) != TYPE_NAME(base))
		return false;

	/* Apparently this is needed for Objective-C. */
	if (TYPE_CONTEXT(cand) == TYPE_CONTEXT(base))
		return false;

	/* Check alignment. */
	if (TYPE_ALIGN(cand) != align)
		return false;
	if (!attribute_list_equal(TYPE_ATTRIBUTES(cand), TYPE_ATTRIBUTES(base)))
		return false;

	return true;
}

static inline tree build_aligned_type(tree type, unsigned int align)
{
	tree t;

	if (TYPE_PACKED(type) || TYPE_ALIGN(type) == align)
		return type;

	for (t = TYPE_MAIN_VARIANT(type); t; t = TYPE_NEXT_VARIANT(t))
		if (check_aligned_type(t, type, align))
			return t;

	t = build_variant_type_copy(type);
	TYPE_ALIGN(t) = align;

	return t;
}
#endif

#if BUILDING_GCC_VERSION <= 4006
#define ANY_RETURN_P(rtx)		(GET_CODE(rtx) == RETURN)
#define C_DECL_REGISTER(EXP)		DECL_LANG_FLAG_4(EXP)
#define EDGE_PRESERVE			0ULL
#define HOST_WIDE_INT_PRINT_HEX_PURE	"%" HOST_WIDE_INT_PRINT "x"
#define flag_fat_lto_objects		true
#define HOST_WIDE_INT_1			HOST_WIDE_INT_C(1)
#define LOCATION_COLUMN(LOC)((expand_location (LOC)).column)
#define ALLOCA_FUNCTION_CODE_P(FCODE)	((FCODE) == BUILT_IN_ALLOCA)

#define get_random_seed(noinit) ({						\
	unsigned HOST_WIDE_INT seed;						\
	sscanf(get_random_seed(noinit), "%" HOST_WIDE_INT_PRINT "x", &seed);	\
	seed * seed; })

#define int_const_binop(code, arg1, arg2)	\
	int_const_binop((code), (arg1), (arg2), 0)

static inline void remove_call_arg_locations(rtx insn) {}

static inline tree build_array_type_nelts(tree elt_type, unsigned HOST_WIDE_INT nelts)
{
	return build_array_type(elt_type, build_index_type(size_int (nelts - 1)));
}

static inline bool gimple_clobber_p(const_gimple_stmt_ptr s __unused)
{
	return false;
}

static inline bool gimple_asm_clobbers_memory_p(const_gimple_stmt_ptr stmt)
{
	unsigned i;

	for (i = 0; i < gimple_asm_nclobbers(stmt); i++) {
		tree op = gimple_asm_clobber_op(stmt, i);

		if (!strcmp(TREE_STRING_POINTER(TREE_VALUE(op)), "memory"))
			return true;
	}

	return false;
}

static inline bool gimple_call_internal_p(const_gimple_stmt_ptr gs)
{
	GIMPLE_CHECK(gs, GIMPLE_CALL);
	return false;
}

enum internal_fn { IFN_LAST };

static inline const char *internal_fn_name(enum internal_fn fn)
{
	gcc_unreachable();
}

static inline enum internal_fn gimple_call_internal_fn(const_gimple_stmt_ptr gs)
{
	gcc_unreachable();
}

static inline tree gimple_call_fntype(const_gimple_stmt_ptr call)
{
	tree fntype;

	if (gimple_call_internal_p(call))
		return NULL_TREE;

	if (gimple_num_ops(call) < 2)
		return NULL_TREE;

	fntype = TREE_TYPE(gimple_call_fn(call));
	if (!fntype)
		return NULL_TREE;

	return TREE_CODE(fntype) == POINTER_TYPE ? TREE_TYPE(fntype) : NULL_TREE;
}

static inline tree builtin_decl_implicit(enum built_in_function fncode)
{
	return implicit_built_in_decls[fncode];
}

static inline int ipa_reverse_postorder(cgraph_node_ptr *order)
{
	return cgraph_postorder(order);
}

static inline cgraph_node_ptr cgraph_create_node(tree decl)
{
	return cgraph_node(decl);
}

static inline cgraph_node_ptr cgraph_get_create_node(tree decl)
{
	cgraph_node_ptr node = cgraph_get_node(decl);

	return node ? node : cgraph_node(decl);
}

static inline bool cgraph_function_with_gimple_body_p(cgraph_node_ptr node)
{
	return node->analyzed && !node->thunk.thunk_p && !node->alias;
}

static inline cgraph_node_ptr cgraph_first_function_with_gimple_body(void)
{
	cgraph_node_ptr node;

	for (node = cgraph_nodes; node; node = node->next)
		if (cgraph_function_with_gimple_body_p(node))
			return node;
	return NULL;
}

static inline cgraph_node_ptr cgraph_next_function_with_gimple_body(cgraph_node_ptr node)
{
	for (node = node->next; node; node = node->next)
		if (cgraph_function_with_gimple_body_p(node))
			return node;
	return NULL;
}

/* Given NODE, walk the alias chain to return the function NODE is alias of.
   Walk through thunk, too.
   When AVAILABILITY is non-NULL, get minimal availablity in the chain.  */
static inline struct cgraph_node *cgraph_function_node(struct cgraph_node *node, enum availability *availability)
{
	if (availability)
		*availability = cgraph_function_body_availability(node);
	while (node) {
		if (node->thunk.thunk_p)
			node = node->callees->callee;
		else
			return node;
		if (availability) {
			enum availability a = cgraph_function_body_availability(node);
			if (a < *availability)
				*availability = a;
		}
	}
	return NULL;
}

static inline bool cgraph_for_node_and_aliases(cgraph_node_ptr node, bool (*callback)(cgraph_node_ptr, void *), void *data, bool include_overwritable)
{
	cgraph_node_ptr alias;

	if (callback(node, data))
		return true;

	for (alias = node->same_body; alias; alias = alias->next) {
		if (include_overwritable || cgraph_function_body_availability(alias) > AVAIL_OVERWRITABLE)
			if (cgraph_for_node_and_aliases(alias, callback, data, include_overwritable))
				return true;
	}

	return false;
}

#define FOR_EACH_FUNCTION_WITH_GIMPLE_BODY(node) \
	for ((node) = cgraph_first_function_with_gimple_body(); (node); \
		(node) = cgraph_next_function_with_gimple_body(node))

static inline bool cgraph_edge_recursive_p(cgraph_edge_p edge)
{
	return edge->callee->decl == (edge->caller->global.inlined_to ? : edge->caller)->decl;
}

static inline void varpool_add_new_variable(tree decl)
{
	varpool_finalize_decl(decl);
}

static inline source_location expansion_point_location(source_location location)
{
	return location;
}
#endif

#if BUILDING_GCC_VERSION == 4006 || BUILDING_GCC_VERSION == 4007
#define ipa_ref_list_referring_iterate(L, I, P) ipa_ref_list_refering_iterate((L), (I), (P))

static inline bool get_ptr_info_alignment(struct ptr_info_def *pi, unsigned int *alignp, unsigned int *misalignp)
{
	*alignp = pi->align;
	*misalignp = pi->misalign;
	return true;
}

static inline cgraph_node_ptr ipa_ref_referring_cgraph_node(struct ipa_ref *ref)
{
	return ref->refering_type == IPA_REF_CGRAPH ? ipa_ref_refering_node(ref) : NULL;
}

static inline varpool_node_ptr ipa_ref_referring_varpool_node(struct ipa_ref *ref)
{
	
	return ref->refering_type == IPA_REF_VARPOOL ? ipa_ref_refering_varpool_node(ref) : NULL;
}
#endif

#if BUILDING_GCC_VERSION <= 4007
#define FOR_EACH_FUNCTION(node)	\
	for (node = cgraph_nodes; node; node = node->next)
#define FOR_EACH_VARIABLE(node)	\
	for (node = varpool_nodes; node; node = node->next)
#define PROP_loops 0
#define NODE_SYMBOL(node) (node)
#define NODE_DECL(node) (node)->decl
#define INSN_LOCATION(INSN) INSN_LOCATOR(INSN)
#define SET_INSN_LOCATION(INSN, LOC) INSN_LOCATOR(INSN) = (LOC)
#define vNULL NULL
#define FUNCTION_POINTER_TYPE_P(TYPE) (POINTER_TYPE_P(TYPE) && TREE_CODE(TREE_TYPE(TYPE)) == FUNCTION_TYPE)
#define dump_bb(outf, bb, indent, flags) dump_bb((bb), (outf), (indent))
#define print_rtl_with_bb(outf, first, flags)			\
do {								\
	if (flags & TDF_SLIM)					\
		print_rtl_slim_with_bb(outf, first, flags);	\
	else							\
		print_rtl_with_bb(outf, first);			\
} while (0)
#define varpool_node_for_decl(decl) varpool_node(decl)

// unfortunately gimple_phi_set_result (also used by create_phi_node)
// in earlier gcc versions doesn't do these...
#define gimple_phi_set_result(phi, result)			\
do {								\
	gimple_phi_set_result(phi, result);			\
	if (result == NULL_TREE)				\
		break;						\
	if (TREE_CODE(result) != SSA_NAME)			\
		break;						\
	SSA_NAME_DEF_STMT(result) = phi;			\
} while (0)
#define create_phi_node(var, bb) ({				\
	gphi *phi = create_phi_node((var), (bb));		\
								\
	SSA_NAME_DEF_STMT(gimple_phi_result(phi)) = phi;	\
	phi;							\
})

#ifndef __cplusplus
#define gimple_clobber_p(s) gimple_clobber_p(CONST_CAST_GIMPLE(s))
#define gimple_assign_cast_p(s) gimple_assign_cast_p(CONST_CAST_GIMPLE(s))
#define gimple_assign_single_p(s) gimple_assign_single_p(CONST_CAST_GIMPLE(s))
#endif

#if HOST_BITS_PER_LONG >= 64 || !defined NEED_64BIT_HOST_WIDE_INT
# define HOST_WIDE_INT_C(X) X ## L
#else
# if HOST_BITS_PER_LONGLONG >= 64
#  define HOST_WIDE_INT_C(X) X ## LL
# else
#  if HOST_BITS_PER___INT64 >= 64
#   define HOST_WIDE_INT_C(X) X ## i64
#  else
    #error "Unable to find a suitable type for HOST_WIDE_INT"
#  endif
# endif
#endif

extern struct opt_pass **passes_by_id;
extern int passes_by_id_size;

static inline int bb_loop_depth(const_basic_block bb)
{
	return bb->loop_father ? loop_depth(bb->loop_father) : 0;
}

static inline bool gimple_store_p(gimple_stmt_ptr gs)
{
	tree lhs = gimple_get_lhs(gs);

	return lhs && !is_gimple_reg(lhs);
}

static inline void gimple_init_singleton(gimple_stmt_ptr g __unused)
{
}

static inline gimple_stmt_iterator gsi_none(void)
{
	gimple_stmt_iterator i;
	i.ptr = NULL;
	i.seq = NULL;
	i.bb = NULL;
	return i;
}

static inline tree copy_ssa_name(tree var, gimple_stmt_ptr stmt)
{
	return make_ssa_name(SSA_NAME_VAR(var), stmt);
}

static inline tree make_temp_ssa_name(tree type, gimple_stmt_ptr stmt, const char *name)
{
	tree var;

	var = create_tmp_var(type, name);
	add_referenced_var(var);
	return make_ssa_name(var, stmt);
}

static inline gimple_stmt_ptr build_type_cast(tree to_type, tree op, tree lhs)
{
	if (lhs == NULL_TREE)
		lhs = make_temp_ssa_name(to_type, NULL, NULL);
	return gimple_build_assign_with_ops(NOP_EXPR, lhs, op, NULL_TREE);
}

static inline struct loop *loop_outermost(struct loop *loop)
{
	return VEC_length(loop_p, loop->superloops) <= 1 ? loop : VEC_index(loop_p, loop->superloops, 1);
}

#define SBITMAP_SIZE(SBITMAP) ((SBITMAP)->n_bits)

static inline SBITMAP_ELT_TYPE sbitmap_bit_p(const_sbitmap map, unsigned int bitno)
{
	return TEST_BIT(map, bitno);
}

static inline void sbitmap_set_bit(sbitmap map, unsigned int bitno)
{
	SET_BIT(map, bitno);
}

static inline void sbitmap_clear_bit(sbitmap map, unsigned int bitno)
{
	RESET_BIT(map, bitno);
}

static inline void sbitmap_clear(sbitmap map)
{
	sbitmap_zero(map);
}

static inline unsigned int sbitmap_count_bits(const_sbitmap map)
{
	unsigned int count = 0;

	for (unsigned int i = 0; i < map->size; i++)
		if (map->elms[i]) {
#if HOST_BITS_PER_WIDEST_FAST_INT == HOST_BITS_PER_LONG
		count += __builtin_popcountl(map->elms[i]);
#elif HOST_BITS_PER_WIDEST_FAST_INT == HOST_BITS_PER_LONGLONG
		count += __builtin_popcountll(map->elms[i]);
#else
		count += __builtin_popcount(map->elms[i]);
#endif
		}

	return count;
}

static inline bool sbitmap_and(sbitmap dst, const_sbitmap a, const_sbitmap b)
{
	return sbitmap_a_and_b_cg(dst, a, b);
}

static inline bool sbitmap_ior(sbitmap dst, const_sbitmap a, const_sbitmap b)
{
	return sbitmap_a_or_b_cg(dst, a, b);
}

static inline void set_ssa_default_def(struct function *fn, tree var, tree def)
{
	if (gimple_default_def(fn, var) != NULL_TREE)
		set_default_def(var, def);
}
#endif

#if BUILDING_GCC_VERSION <= 4007 || (BUILDING_GCC_VERSION == 4008 && BUILDING_GCC_PATCHLEVEL < 5)
enum expand_modifier {
	EXPAND_NORMAL = 0,
	EXPAND_STACK_PARM,
	EXPAND_SUM,
	EXPAND_CONST_ADDRESS,
	EXPAND_INITIALIZER,
	EXPAND_WRITE,
	EXPAND_MEMORY
};

rtx expand_expr_real(tree, rtx, enum machine_mode, enum expand_modifier, rtx *);

static inline rtx expand_expr(tree exp, rtx target, enum machine_mode mode, enum expand_modifier modifier)
{
	return expand_expr_real(exp, target, mode, modifier, NULL);
}
#endif

#if BUILDING_GCC_VERSION == 4007 || BUILDING_GCC_VERSION == 4008
static inline cgraph_node_ptr cgraph_alias_target(cgraph_node_ptr n)
{
	return cgraph_alias_aliased_node(n);
}
#endif

#if BUILDING_GCC_VERSION >= 4007 && BUILDING_GCC_VERSION <= 4009
#define cgraph_create_edge(caller, callee, call_stmt, count, freq, nest) \
	cgraph_create_edge((caller), (callee), (call_stmt), (count), (freq))
#define cgraph_create_edge_including_clones(caller, callee, old_call_stmt, call_stmt, count, freq, nest, reason) \
	cgraph_create_edge_including_clones((caller), (callee), (old_call_stmt), (call_stmt), (count), (freq), (reason))
#define FOR_EACH_ALIAS(node, ref) \
	for (unsigned int ref##_idx = 0; ipa_ref_list_referring_iterate(&NODE_SYMBOL(node)->ref_list, ref##_idx, ref); ref##_idx++)

#ifdef __cplusplus
static inline bool gimple_clobber_p(const_gimple_stmt_ptr s)
{
	return gimple_clobber_p(CONST_CAST_GIMPLE(s));
}

static inline bool gimple_assign_cast_p(const_gimple_stmt_ptr s)
{
	return gimple_assign_cast_p(CONST_CAST_GIMPLE(s));
}

static inline bool gimple_assign_single_p(const_gimple_stmt_ptr s)
{
	return gimple_assign_single_p(CONST_CAST_GIMPLE(s));
}
#else
#define gimple_clobber_p(s) gimple_clobber_p(CONST_CAST_GIMPLE(s))
#endif
#endif

#if BUILDING_GCC_VERSION >= 4007
#define get_pointer_alignment(exp, max_align) get_pointer_alignment(exp)
#endif

#if BUILDING_GCC_VERSION <= 4008
#define ENTRY_BLOCK_PTR_FOR_FN(FN)	ENTRY_BLOCK_PTR_FOR_FUNCTION(FN)
#define EXIT_BLOCK_PTR_FOR_FN(FN)	EXIT_BLOCK_PTR_FOR_FUNCTION(FN)
#define basic_block_info_for_fn(FN)	((FN)->cfg->x_basic_block_info)
#define n_basic_blocks_for_fn(FN)	((FN)->cfg->x_n_basic_blocks)
#define n_edges_for_fn(FN)		((FN)->cfg->x_n_edges)
#define last_basic_block_for_fn(FN)	((FN)->cfg->x_last_basic_block)
#define label_to_block_map_for_fn(FN)	((FN)->cfg->x_label_to_block_map)
#define profile_status_for_fn(FN)	((FN)->cfg->x_profile_status)
#define BASIC_BLOCK_FOR_FN(FN, N)	BASIC_BLOCK_FOR_FUNCTION((FN), (N))
#define NODE_IMPLICIT_ALIAS(node)	(node)->same_body_alias
#define VAR_P(NODE)			(TREE_CODE(NODE) == VAR_DECL)
#ifndef HOST_WIDE_INT_UC
#define HOST_WIDE_INT_UC(X)		HOST_WIDE_INT_C(X ## U)
#endif
#ifndef HOST_WIDE_INT_1U
#define HOST_WIDE_INT_1U		HOST_WIDE_INT_UC(1)
#define EXPR_LOC_OR_LOC(NODE, LOCUS)	(EXPR_HAS_LOCATION(NODE) ? (NODE)->exp.locus : (LOCUS))
#endif

#define REG_BR_PROB_NEVER		0
#define REG_BR_PROB_VERY_UNLIKELY	(REG_BR_PROB_BASE / 2000 - 1)
#define REG_BR_PROB_UNLIKELY		(REG_BR_PROB_BASE / 5 - 1)
#define REG_BR_PROB_EVEN		(REG_BR_PROB_BASE / 2)
#define REG_BR_PROB_LIKELY		(REG_BR_PROB_BASE - REG_BR_PROB_UNLIKELY)
#define REG_BR_PROB_VERY_LIKELY		(REG_BR_PROB_BASE - REG_BR_PROB_VERY_UNLIKELY)
#define REG_BR_PROB_ALWAYS		(REG_BR_PROB_BASE)

static inline bool recursive_call_p(tree func, tree dest)
{
	return func == dest;
}

static inline bool tree_fits_shwi_p(const_tree t)
{
	if (t == NULL_TREE || TREE_CODE(t) != INTEGER_CST)
		return false;

	if (TREE_INT_CST_HIGH(t) == 0 && (HOST_WIDE_INT)TREE_INT_CST_LOW(t) >= 0)
		return true;

	if (TREE_INT_CST_HIGH(t) == -1 && (HOST_WIDE_INT)TREE_INT_CST_LOW(t) < 0 && !TYPE_UNSIGNED(TREE_TYPE(t)))
		return true;

	return false;
}

static inline bool tree_fits_uhwi_p(const_tree t)
{
	if (t == NULL_TREE || TREE_CODE(t) != INTEGER_CST)
		return false;

	return TREE_INT_CST_HIGH(t) == 0;
}

static inline HOST_WIDE_INT tree_to_shwi(const_tree t)
{
	gcc_assert(tree_fits_shwi_p(t));
	return TREE_INT_CST_LOW(t);
}

static inline unsigned HOST_WIDE_INT tree_to_uhwi(const_tree t)
{
	gcc_assert(tree_fits_uhwi_p(t));
	return TREE_INT_CST_LOW(t);
}

static inline const char *get_tree_code_name(enum tree_code code)
{
	gcc_assert(code < MAX_TREE_CODES);
	return tree_code_name[code];
}

static inline HOST_WIDE_INT sext_hwi(HOST_WIDE_INT src, unsigned int prec)
{
	if (prec == HOST_BITS_PER_WIDE_INT)
		return src;

	if (prec < HOST_BITS_PER_WIDE_INT) {
#if defined(__GNUC__)
		int shift = HOST_BITS_PER_WIDE_INT - prec;
		return ((HOST_WIDE_INT)((unsigned HOST_WIDE_INT)src << shift)) >> shift;
#else
		HOST_WIDE_INT sign_mask = HOST_WIDE_INT_1 << (prec - 1);
		HOST_WIDE_INT value_mask = (HOST_WIDE_INT_1U << prec) - HOST_WIDE_INT_1U;
		return (((src & value_mask) ^ sign_mask) - sign_mask);
#endif
	}

	gcc_unreachable();
}

static inline HOST_WIDE_INT zext_hwi(HOST_WIDE_INT src, unsigned int prec)
{
	if (prec == HOST_BITS_PER_WIDE_INT)
		return src;

	if (prec < HOST_BITS_PER_WIDE_INT)
		return src & ((HOST_WIDE_INT_1U << prec) - 1);

	gcc_unreachable();
}

static inline void swap_ssa_operands(gimple_stmt_ptr stmt, tree *exp0, tree *exp1)
{
	swap_tree_operands(stmt, exp0, exp1);
}

static inline void gimple_seq_set_location(gimple_seq seq, location_t loc)
{
	gimple_stmt_iterator i;

	for (i = gsi_start(seq); !gsi_end_p(i); gsi_next(&i))
		gimple_set_location(gsi_stmt(i), loc);
}

#define ipa_remove_stmt_references(cnode, stmt)

typedef union gimple_statement_d gasm;
typedef union gimple_statement_d gassign;
typedef union gimple_statement_d gcall;
typedef union gimple_statement_d gcond;
typedef union gimple_statement_d gdebug;
typedef union gimple_statement_d ggoto;
typedef union gimple_statement_d gphi;
typedef union gimple_statement_d greturn;
typedef union gimple_statement_d gswitch;
typedef union gimple_statement_d glabel;

typedef gimple_stmt_iterator gphi_iterator;

static inline gasm *as_a_gasm(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gasm *as_a_const_gasm(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gassign *as_a_gassign(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gassign *as_a_const_gassign(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gcall *as_a_gcall(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gcall *as_a_const_gcall(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gcond *as_a_gcond(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gcond *as_a_const_gcond(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gdebug *as_a_gdebug(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gdebug *as_a_const_gdebug(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline ggoto *as_a_ggoto(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const ggoto *as_a_const_ggoto(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gphi *as_a_gphi(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gphi *as_a_const_gphi(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline greturn *as_a_greturn(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const greturn *as_a_const_greturn(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gswitch *as_a_gswitch(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gswitch *as_a_const_gswitch(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gphi* gsi_gphi(gphi_iterator gsi)
{
	return as_a_gphi(gsi_stmt(gsi));
}

static inline const glabel *as_a_const_glabel(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gimple_stmt_iterator gsi_start_nondebug_after_labels_bb(basic_block bb)
{
	gimple_stmt_iterator i = gsi_after_labels(bb);

	if (!gsi_end_p(i) && is_gimple_debug(gsi_stmt(i)))
		gsi_next_nondebug(&i);

	return i;
}

static bool __pointer_set_elements(const void *elem, void *data)
{
	size_t *count = (size_t *)data;

	(*count)++;
	return false;
}

static inline size_t pointer_set_elements(const struct pointer_set_t *set)
{
	size_t count = 0;

	pointer_set_traverse(set, __pointer_set_elements, &count);
	return count;
}
#endif

#if BUILDING_GCC_VERSION == 4008
#define NODE_SYMBOL(node) (&(node)->symbol)
#define NODE_DECL(node) (node)->symbol.decl

static inline gimple_stmt_ptr build_type_cast(tree to_type, tree op, tree lhs)
{
	if (lhs == NULL_TREE)
		lhs = make_ssa_name(to_type, NULL);
	return gimple_build_assign_with_ops(NOP_EXPR, lhs, op, NULL_TREE);
}
#endif

#if BUILDING_GCC_VERSION >= 4008
#define add_referenced_var(var)
#define mark_sym_for_renaming(var)
#define varpool_mark_needed_node(node)
#define create_var_ann(var)
#define TODO_dump_func 0
#define TODO_dump_cgraph 0

#define VEC_index(T, V, I) ((*(V))[(I)])
#define VEC_replace(T, V, I, O) (*(V))[(I)] = (O)
#define VEC_safe_push(T, A, V, O) vec_safe_push((V), (O))

#define INSN_LOCATOR(INSN) INSN_LOCATION(INSN)
#define SET_INSN_LOCATION(INSN, LOC) INSN_LOCATION(INSN) = (LOC)

#define EXECUTE_IF_SET_IN_SBITMAP(BITMAP, MIN, BITNUM, ITER)		\
	for (bmp_iter_set_init(&(ITER), (BITMAP), (MIN), &(BITNUM));	\
		bmp_iter_set (&(ITER), &(BITNUM));			\
		bmp_iter_next (&(ITER), &(BITNUM)))

static inline SBITMAP_ELT_TYPE sbitmap_bit_p(const_sbitmap map, unsigned int bitno)
{
	return bitmap_bit_p(map, bitno);
}

static inline void sbitmap_set_bit(sbitmap map, unsigned int bitno)
{
	bitmap_set_bit(map, bitno);
}

static inline void sbitmap_clear_bit(sbitmap map, unsigned int bitno)
{
	bitmap_clear_bit(map, bitno);
}

static inline void sbitmap_clear(sbitmap map)
{
	bitmap_clear(map);
}

static inline void sbitmap_ones(sbitmap map)
{
	bitmap_ones(map);
}

static inline int sbitmap_first_set_bit(const_sbitmap map)
{
	return bitmap_first_set_bit(map);
}

static inline void debug_sbitmap(const_sbitmap map)
{
	debug_bitmap(map);
}

static inline bool sbitmap_empty_p(const_sbitmap map)
{
	return bitmap_empty_p(map);
}

static inline unsigned int sbitmap_count_bits(const_sbitmap map)
{
	unsigned int count = 0;

	for (unsigned int i = 0; i < map->size; i++)
		if (map->elms[i]) {
#if HOST_BITS_PER_WIDEST_FAST_INT == HOST_BITS_PER_LONG
		count += __builtin_popcountl(map->elms[i]);
#elif HOST_BITS_PER_WIDEST_FAST_INT == HOST_BITS_PER_LONGLONG
		count += __builtin_popcountll(map->elms[i]);
#else
		count += __builtin_popcount(map->elms[i]);
#endif
		}

	return count;
}

static inline void sbitmap_copy(sbitmap dst, const_sbitmap src)
{
	memcpy(dst->elms, src->elms, sizeof(SBITMAP_ELT_TYPE) * dst->size);
}

static inline int sbitmap_first_set_bit(const sbitmap map)
{
	return bitmap_first_set_bit(map);
}

static inline int sbitmap_last_set_bit(const sbitmap map)
{
	return bitmap_last_set_bit(map);
}

static inline bool sbitmap_and(sbitmap dst, const_sbitmap a, const_sbitmap b)
{
	return bitmap_and(dst, a, b);
}

static inline bool sbitmap_ior(sbitmap dst, const_sbitmap a, const_sbitmap b)
{
	return bitmap_ior(dst, a, b);
}
#endif

static inline basic_block nearest_common_dominator_for_sbitmap(enum cdi_direction dir, const_sbitmap blocks)
{
	unsigned int i;
	sbitmap_iterator sbi;
	basic_block dom;

	if (-1 == sbitmap_first_set_bit(blocks))
		return NULL;

	dom = BASIC_BLOCK_FOR_FN(cfun, sbitmap_first_set_bit(blocks));
	EXECUTE_IF_SET_IN_SBITMAP(blocks, 0, i, sbi)
		if (dom != BASIC_BLOCK_FOR_FN(cfun, i))
			dom = nearest_common_dominator(dir, dom, BASIC_BLOCK_FOR_FN(cfun, i));

	return dom;
}

#if BUILDING_GCC_VERSION == 4008 || BUILDING_GCC_VERSION == 4009
#define ipa_ref_referring_cgraph_node(ref) ({ \
	NODE_SYMBOL(ref->referring)->type == SYMTAB_FUNCTION ? ipa_ref_referring_node(ref) : NULL; \
})

#define ipa_ref_referring_varpool_node(ref) ({ \
	NODE_SYMBOL(ref->referring)->type == SYMTAB_VARIABLE ? ipa_ref_referring_varpool_node(ref) : NULL; \
})
#endif

#if (BUILDING_GCC_VERSION == 4008 && BUILDING_GCC_PATCHLEVEL == 5) || BUILDING_GCC_VERSION == 4009
enum expand_modifier {
	EXPAND_NORMAL = 0,
	EXPAND_STACK_PARM,
	EXPAND_SUM,
	EXPAND_CONST_ADDRESS,
	EXPAND_INITIALIZER,
	EXPAND_WRITE,
	EXPAND_MEMORY
};

rtx expand_expr_real(tree, rtx, enum machine_mode, enum expand_modifier, rtx *, bool);

static inline rtx expand_expr(tree exp, rtx target, enum machine_mode mode, enum expand_modifier modifier)
{
	return expand_expr_real(exp, target, mode, modifier, NULL, false);
}
#endif

#if BUILDING_GCC_VERSION >= 4008 && BUILDING_GCC_VERSION < 11000
#define VEC(T, A) vec<T, va_##A>
#endif

#if BUILDING_GCC_VERSION <= 4009
#define TODO_verify_il 0
#define AVAIL_INTERPOSABLE AVAIL_OVERWRITABLE

#define section_name_prefix LTO_SECTION_NAME_PREFIX
#define fatal_error(loc, gmsgid, ...) fatal_error((gmsgid), __VA_ARGS__)

typedef struct pointer_set_t gimple_set;
typedef struct pointer_set_t const_gimple_set;
typedef struct pointer_set_t tree_set;
typedef struct pointer_set_t const_tree_set;
typedef struct pointer_set_t cgraph_set;

static inline gimple_set *gimple_pointer_set_create(void)
{
	return pointer_set_create();
}

static inline const_gimple_set *const_gimple_pointer_set_create(void)
{
	return pointer_set_create();
}

static inline tree_set *tree_pointer_set_create(void)
{
	return pointer_set_create();
}

static inline const_tree_set *const_tree_pointer_set_create(void)
{
	return pointer_set_create();
}

static inline cgraph_set *cgraph_pointer_set_create(void)
{
	return pointer_set_create();
}

#define gt_ggc_mx_hash_set_tree_ gt_ggc_mx_pointer_map_t
#define gt_pch_nx_hash_set_tree_ gt_pch_nx_pointer_map_t

typedef struct pointer_map_t tree_tree_map;

static inline tree_tree_map *tree_tree_map_create(void)
{
	return pointer_map_create();
}

static inline bool tree_tree_map_insert(tree_tree_map *map, tree key, tree value)
{
	tree *slot;

	gcc_assert(value != NULL_TREE);
	slot = (tree *)pointer_map_insert(map, key);
	gcc_assert(slot);
	if (*slot != NULL_TREE)
		return true;
	*slot = value;
	return false;
}

static inline tree *tree_tree_map_get(tree_tree_map *map, tree key)
{
	tree *slot;

	slot = (tree *)pointer_map_contains(map, key);
	return slot ? slot : 0;
}

static inline void tree_tree_map_traverse(tree_tree_map *map, bool (*fn)(const void *, void **, void *), void *data)
{
	pointer_map_traverse(map, fn, data);
}

static bool __tree_tree_map_elements(const void *key, void **value, void *data)
{
	size_t *count = (size_t *)data;

	(*count)++;
	return false;
}

static inline size_t tree_tree_map_elements(const struct pointer_map_t *map)
{
	size_t count = 0;

	pointer_map_traverse(map, __tree_tree_map_elements, &count);
	return count;
}

static inline bool tree_tree_map_contains(tree_tree_map *map, tree key)
{
	return !!pointer_map_contains(map, key);
}

static inline void tree_tree_map_destroy(tree_tree_map *map)
{
	pointer_map_destroy(map);
}

#define gt_ggc_mx_hash_map_tree_tree_ gt_ggc_mx_pointer_map_t
#define gt_pch_nx_hash_map_tree_tree_ gt_pch_nx_pointer_map_t

rtx emit_move_insn(rtx x, rtx y);
rtx copy_addr_to_reg(rtx);
bool df_insn_rescan(rtx insn);

typedef struct rtx_def rtx_insn;
typedef rtx_insn rtx_code_label;

static inline const char *get_decl_section_name(const_tree decl)
{
	if (DECL_SECTION_NAME(decl) == NULL_TREE)
		return NULL;

	return TREE_STRING_POINTER(DECL_SECTION_NAME(decl));
}

static inline void set_decl_section_name(tree node, const char *value)
{
	DECL_ATTRIBUTES(node) = remove_attribute("section", DECL_ATTRIBUTES(node));
	if (value) {
		tree section = build_const_char_string(strlen(value) + 1, value);

		DECL_SECTION_NAME(node) = section;
		section = tree_cons(NULL_TREE, section, NULL_TREE);
		DECL_ATTRIBUTES(node) = tree_cons(get_identifier("section"), section, DECL_ATTRIBUTES(node));
	} else
		DECL_SECTION_NAME(node) = NULL_TREE;
}

static inline varpool_node_ptr varpool_get_for_asmname(tree asmname)
{
	return varpool_node_for_asm(asmname);
}

static inline basic_block insert_cond_bb(basic_block cond_bb, gimple_stmt_ptr stmt, gimple_stmt_ptr cond, int prob)
{
	edge true_edge, false_edge;
	gimple_stmt_iterator gsi;
	basic_block join_bb, true_bb;

	false_edge = split_block(cond_bb, stmt);
	gcc_assert(cond_bb == false_edge->src);
	join_bb = false_edge->dest;
	false_edge->flags = EDGE_FALSE_VALUE;
	false_edge->probability -= prob;

	gcc_assert(gimple_code(cond) == GIMPLE_COND);
	gsi = gsi_last_bb(cond_bb);
	if (gsi_end_p(gsi))
		gsi_insert_before(&gsi, cond, GSI_CONTINUE_LINKING);
	else
		gsi_insert_after(&gsi, cond, GSI_CONTINUE_LINKING);

	true_bb = create_empty_bb(cond_bb);
	true_edge = make_edge(cond_bb, true_bb, EDGE_TRUE_VALUE);
	true_edge->probability = prob;
	true_bb->count = true_edge->count;

	make_single_succ_edge(true_bb, join_bb, EDGE_FALLTHRU);

	if (dom_info_available_p(CDI_DOMINATORS)) {
		set_immediate_dominator(CDI_DOMINATORS, true_bb, cond_bb);
		set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);
	}

	if (current_loops) {
		add_bb_to_loop(true_bb, cond_bb->loop_father);
		loops_state_set(LOOPS_NEED_FIXUP);
	}

	return true_bb;
}

typedef unsigned short poly_uint16_pod;
typedef HOST_WIDE_INT poly_int64_pod;
typedef unsigned HOST_WIDE_INT poly_uint64_pod;
//typedef offset_int poly_offset_int_pod;
//typedef wide_int poly_wide_int_pod;
//typedef widest_int poly_widest_int_pod;

typedef unsigned short poly_uint16;
typedef HOST_WIDE_INT poly_int64;
typedef unsigned HOST_WIDE_INT poly_uint64;
//typedef offset_int poly_offset_int;
//typedef wide_int poly_wide_int;
//typedef wide_int_ref poly_wide_int_ref;
//typedef widest_int poly_widest_int;

typedef double_int wide_int;
#endif

#if BUILDING_GCC_VERSION == 4009
#define pointer_set_elements(set) set->n_elements
#define FOR_EACH_INSN_USE(ITER, INSN) for (ITER = DF_INSN_USES(INSN); *ITER; ITER++)

typedef struct gimple_statement_asm gasm;
typedef struct gimple_statement_base gassign;
typedef struct gimple_statement_call gcall;
typedef struct gimple_statement_base gcond;
typedef struct gimple_statement_base gdebug;
typedef struct gimple_statement_base ggoto;
typedef struct gimple_statement_phi gphi;
typedef struct gimple_statement_base greturn;
typedef struct gimple_statement_base gswitch;
typedef struct gimple_statement_base glabel;

typedef gimple_stmt_iterator gphi_iterator;

static inline gasm *as_a_gasm(gimple_stmt_ptr stmt)
{
	return as_a<gasm>(stmt);
}

static inline const gasm *as_a_const_gasm(const_gimple_stmt_ptr stmt)
{
	return as_a<const gasm>(stmt);
}

static inline gassign *as_a_gassign(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gassign *as_a_const_gassign(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gcall *as_a_gcall(gimple_stmt_ptr stmt)
{
	return as_a<gcall>(stmt);
}

static inline const gcall *as_a_const_gcall(const_gimple_stmt_ptr stmt)
{
	return as_a<const gcall>(stmt);
}

static inline gcond *as_a_gcond(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gcond *as_a_const_gcond(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gdebug *as_a_gdebug(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gdebug *as_a_const_gdebug(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline ggoto *as_a_ggoto(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const ggoto *as_a_const_ggoto(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gphi *as_a_gphi(gimple_stmt_ptr stmt)
{
	return as_a<gphi>(stmt);
}

static inline const gphi *as_a_const_gphi(const_gimple_stmt_ptr stmt)
{
	return as_a<const gphi>(stmt);
}

static inline greturn *as_a_greturn(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const greturn *as_a_const_greturn(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gswitch *as_a_gswitch(gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const gswitch *as_a_const_gswitch(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline const glabel *as_a_const_glabel(const_gimple_stmt_ptr stmt)
{
	return stmt;
}

static inline gphi* gsi_gphi(gphi_iterator gsi)
{
	return as_a_gphi(gsi_stmt(gsi));
}
#endif

#if BUILDING_GCC_VERSION >= 4009
#define TODO_ggc_collect 0
#define NODE_SYMBOL(node) (node)
#define NODE_DECL(node) (node)->decl
#define cgraph_node_name(node) (node)->name()
#define NODE_IMPLICIT_ALIAS(node) (node)->cpp_implicit_alias

static inline opt_pass *get_pass_for_id(int id)
{
	return g->get_passes()->get_pass_for_id(id);
}
#endif

#if BUILDING_GCC_VERSION >= 5000
#define TODO_verify_ssa TODO_verify_il
#define TODO_verify_flow TODO_verify_il
#define TODO_verify_stmts TODO_verify_il
#define TODO_verify_rtl_sharing TODO_verify_il

//#define TREE_INT_CST_HIGH(NODE) ({ TREE_INT_CST_EXT_NUNITS(NODE) > 1 ? (unsigned HOST_WIDE_INT)TREE_INT_CST_ELT(NODE, 1) : 0; })

#define INSN_DELETED_P(insn) (insn)->deleted()

static inline gphi *gsi_gphi(gphi_iterator gsi)
{
	return gsi.phi();
}

static inline const char *get_decl_section_name(const_tree decl)
{
	return DECL_SECTION_NAME(decl);
}

#define set_decl_section_name(_node, _value)						\
do {											\
	tree node = (_node);								\
	const char *value = (_value);							\
											\
	DECL_ATTRIBUTES(node) = remove_attribute("section", DECL_ATTRIBUTES(node));	\
	set_decl_section_name(node, value);						\
	if (value) {									\
		tree section = build_const_char_string(strlen(value) + 1, value);	\
											\
		section = tree_cons(NULL_TREE, section, NULL_TREE);			\
		DECL_ATTRIBUTES(node) = tree_cons(get_identifier("section"), section, DECL_ATTRIBUTES(node)); \
	}										\
} while (0)

/* symtab/cgraph related */
#define debug_cgraph_node(node) (node)->debug()
#define cgraph_get_node(decl) cgraph_node::get(decl)
#define cgraph_get_create_node(decl) cgraph_node::get_create(decl)
#define cgraph_create_node(decl) cgraph_node::create(decl)
#define cgraph_node_for_asm(asmname) cgraph_node::get_for_asmname(asmname)
#define cgraph_n_nodes symtab->cgraph_count
#define cgraph_max_uid symtab->cgraph_max_uid
#define cgraph_edge_recursive_p(edge) edge->recursive_p()
#define varpool_get_node(decl) varpool_node::get(decl)
#define varpool_get_for_asmname(asmname) varpool_node::get_for_asmname(asmname)
#define varpool_assemble_decl(node) (node)->assemble_decl()
#define dump_varpool_node(file, node) (node)->dump(file)

typedef struct cgraph_node *cgraph_node_ptr;
typedef struct cgraph_edge *cgraph_edge_p;
typedef struct varpool_node *varpool_node_ptr;

static inline void change_decl_assembler_name(tree decl, tree name)
{
	symtab->change_decl_assembler_name(decl, name);
}

static inline void varpool_finalize_decl(tree decl)
{
	varpool_node::finalize_decl(decl);
}

static inline void varpool_add_new_variable(tree decl)
{
	varpool_node::add(decl);
}

static inline unsigned int rebuild_cgraph_edges(void)
{
	return cgraph_edge::rebuild_edges();
}

static inline cgraph_node_ptr cgraph_function_node(cgraph_node_ptr node, enum availability *availability)
{
	return node->function_symbol(availability);
}

static inline cgraph_node_ptr cgraph_function_or_thunk_node(cgraph_node_ptr node, enum availability *availability = NULL)
{
	return node->ultimate_alias_target(availability);
}

static inline bool cgraph_only_called_directly_p(cgraph_node_ptr node)
{
	return node->only_called_directly_p();
}

static inline enum availability cgraph_function_body_availability(cgraph_node_ptr node)
{
	return node->get_availability();
}

static inline cgraph_node_ptr cgraph_alias_target(cgraph_node_ptr node)
{
	return node->get_alias_target();
}

static inline bool cgraph_for_node_and_aliases(cgraph_node_ptr node, bool (*callback)(cgraph_node_ptr, void *), void *data, bool include_overwritable)
{
	return node->call_for_symbol_and_aliases(callback, data, include_overwritable);
}

static inline bool cgraph_for_node_thunks_and_aliases(cgraph_node_ptr node, bool (*callback)(cgraph_node_ptr, void *), void *data, bool include_overwritable)
{
	return node->call_for_symbol_thunks_and_aliases(callback, data, include_overwritable);
}

static inline struct cgraph_node_hook_list *cgraph_add_function_insertion_hook(cgraph_node_hook hook, void *data)
{
	return symtab->add_cgraph_insertion_hook(hook, data);
}

static inline void cgraph_remove_function_insertion_hook(struct cgraph_node_hook_list *entry)
{
	symtab->remove_cgraph_insertion_hook(entry);
}

static inline struct cgraph_node_hook_list *cgraph_add_node_removal_hook(cgraph_node_hook hook, void *data)
{
	return symtab->add_cgraph_removal_hook(hook, data);
}

static inline void cgraph_remove_node_removal_hook(struct cgraph_node_hook_list *entry)
{
	symtab->remove_cgraph_removal_hook(entry);
}

static inline struct cgraph_2node_hook_list *cgraph_add_node_duplication_hook(cgraph_2node_hook hook, void *data)
{
	return symtab->add_cgraph_duplication_hook(hook, data);
}

static inline void cgraph_remove_node_duplication_hook(struct cgraph_2node_hook_list *entry)
{
	symtab->remove_cgraph_duplication_hook(entry);
}

static inline void cgraph_call_node_duplication_hooks(cgraph_node_ptr node, cgraph_node_ptr node2)
{
	symtab->call_cgraph_duplication_hooks(node, node2);
}

static inline void cgraph_call_edge_duplication_hooks(cgraph_edge_p cs1, cgraph_edge_p cs2)
{
	symtab->call_edge_duplication_hooks(cs1, cs2);
}

/* gimple related */
static inline gimple_stmt_ptr gimple_build_assign_with_ops(enum tree_code subcode, tree lhs, tree op1, tree op2 MEM_STAT_DECL)
{
	return gimple_build_assign(lhs, subcode, op1, op2 PASS_MEM_STAT);
}

template <>
template <>
inline bool is_a_helper<rtx_insn *>::test(rtx_insn *insn __unused)
{
	return true;
}

static inline gasm *as_a_gasm(gimple_stmt_ptr stmt)
{
	return as_a<gasm *>(stmt);
}

static inline const gasm *as_a_const_gasm(const_gimple_stmt_ptr stmt)
{
	return as_a<const gasm *>(stmt);
}

static inline gassign *as_a_gassign(gimple_stmt_ptr stmt)
{
	return as_a<gassign *>(stmt);
}

static inline const gassign *as_a_const_gassign(const_gimple_stmt_ptr stmt)
{
	return as_a<const gassign *>(stmt);
}

static inline gcall *as_a_gcall(gimple_stmt_ptr stmt)
{
	return as_a<gcall *>(stmt);
}

static inline const gcall *as_a_const_gcall(const_gimple_stmt_ptr stmt)
{
	return as_a<const gcall *>(stmt);
}

static inline gcond *as_a_gcond(gimple_stmt_ptr stmt)
{
	return as_a<gcond *>(stmt);
}

static inline const gcond *as_a_const_gcond(const_gimple_stmt_ptr stmt)
{
	return as_a<const gcond *>(stmt);
}

static inline ggoto *as_a_ggoto(gimple_stmt_ptr stmt)
{
	return as_a<ggoto *>(stmt);
}

static inline const ggoto *as_a_const_ggoto(const_gimple_stmt_ptr stmt)
{
	return as_a<const ggoto *>(stmt);
}

static inline gphi *as_a_gphi(gimple_stmt_ptr stmt)
{
	return as_a<gphi *>(stmt);
}

static inline const gphi *as_a_const_gphi(const_gimple_stmt_ptr stmt)
{
	return as_a<const gphi *>(stmt);
}

static inline greturn *as_a_greturn(gimple_stmt_ptr stmt)
{
	return as_a<greturn *>(stmt);
}

static inline const greturn *as_a_const_greturn(const_gimple_stmt_ptr stmt)
{
	return as_a<const greturn *>(stmt);
}

static inline gswitch *as_a_gswitch(gimple_stmt_ptr stmt)
{
	return as_a<gswitch *>(stmt);
}

static inline const glabel *as_a_const_glabel(const_gimple_stmt_ptr stmt)
{
	return as_a<const glabel *>(stmt);
}

static inline const gswitch *as_a_const_gswitch(const_gimple_stmt_ptr stmt)
{
	return as_a<const gswitch *>(stmt);
}

/* pointer/hash set related */
#define HASH_SET(hash, type)						\
typedef struct hash_set<type> hash##_set;				\
									\
static inline hash##_set* hash##_pointer_set_create(void)		\
{									\
	return new hash##_set;						\
}									\
									\
static inline hash##_set* hash##_pointer_set_create_ggc(size_t n = 13)	\
{									\
	return hash##_set::create_ggc(n);				\
}									\
									\
static inline void pointer_set_destroy(hash##_set *set)			\
{									\
	delete set;							\
}									\
									\
static inline bool pointer_set_insert(hash##_set *set, type ptr)	\
{									\
	return set->add(ptr);						\
}									\
									\
static inline bool pointer_set_contains(hash##_set *set, type ptr)	\
{									\
	return set->contains(ptr);					\
}

HASH_SET(gimple, gimple_stmt_ptr)
HASH_SET(const_gimple, const_gimple_stmt_ptr)
HASH_SET(tree, tree)
HASH_SET(const_tree, const_tree)
HASH_SET(cgraph, cgraph_node_ptr)

#undef HASH_SET

#define pointer_set_traverse(set, fn, data) set->traverse<void *, fn>(data)
#define pointer_set_elements(set) set->elements()

/* pointer/hash map related */
#define HASH_MAP(hash, key, value)					\
typedef struct hash_map<key, value> hash##_map;				\
									\
static inline hash##_map *hash##_map_create(void)			\
{									\
	return new hash##_map;						\
}									\
									\
static inline hash##_map *hash##_map_create_ggc(size_t n = 13)		\
{									\
	return hash##_map::create_ggc(n);				\
}									\
									\
static inline void hash##_map_destroy(hash##_map *map)			\
{									\
	delete map;							\
}									\
									\
static inline bool hash##_map_insert(hash##_map *map, key ptr1, value ptr2)\
{									\
	return map->put(ptr1, ptr2);					\
}									\
									\
static inline bool hash##_map_contains(hash##_map *map, key ptr)	\
{									\
	return !!map->get(ptr);						\
}									\
									\
static inline value *hash##_map_get(hash##_map *map, key ptr)		\
{									\
	return map->get(ptr);						\
}

HASH_MAP(tree_tree, tree, tree)

#undef HASH_MAP

#define tree_tree_map_traverse(map, fn, data) map->traverse<void *, fn>(data)
#define tree_tree_map_elements(map) map->elements()

/* IPA/LTO related */
#define ipa_ref_list_referring_iterate(L, I, P)	\
	(L)->referring.iterate((I), &(P))
#define ipa_ref_list_reference_iterate(L, I, P)	\
	(L)->reference.iterate((I), &(P))

static inline cgraph_node_ptr ipa_ref_referring_cgraph_node(struct ipa_ref *ref)
{
	return dyn_cast<cgraph_node_ptr>(ref->referring);
}

static inline varpool_node_ptr ipa_ref_referring_varpool_node(struct ipa_ref *ref)
{
	return dyn_cast<varpool_node_ptr>(ref->referring);
}

static inline void ipa_remove_stmt_references(symtab_node *referring_node, gimple_stmt_ptr stmt)
{
	referring_node->remove_stmt_references(stmt);
}
#endif

#if BUILDING_GCC_VERSION >= 5000 && BUILDING_GCC_VERSION < 6000
/* gimple related */
template <>
template <>
inline bool is_a_helper<const gassign *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_ASSIGN;
}

template <>
template <>
inline bool is_a_helper<const gcond *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_COND;
}
#endif

#if BUILDING_GCC_VERSION < 6000
#define get_inner_reference(exp, pbitsize, pbitpos, poffset, pmode, punsignedp, preversep, pvolatilep, keep_aligning)	\
	({ (void)(reversep); get_inner_reference(exp, pbitsize, pbitpos, poffset, pmode, punsignedp, pvolatilep, keep_aligning); })
#define gen_rtx_set(ARG0, ARG1) gen_rtx_SET(VOIDmode, (ARG0), (ARG1))

#ifndef LEGACY_INT_REG_P
#define LEGACY_INT_REG_P(X) (REG_P(X) && LEGACY_INT_REGNO_P(REGNO(X)))
#endif
#ifndef LEGACY_INT_REGNO_P
#define LEGACY_INT_REGNO_P(N) (IN_RANGE((N), AX_REG, SP_REG))
#endif

#ifndef FIRST_INT_REG
#define FIRST_INT_REG		AX_REG
#endif
#ifndef LAST_INT_REG
#define LAST_INT_REG		SP_REG
#endif

#define FUNC_OR_METHOD_TYPE_P(NODE) (TREE_CODE(NODE) == FUNCTION_TYPE || TREE_CODE(NODE) == METHOD_TYPE)
#endif

#if BUILDING_GCC_VERSION >= 6000
#define gen_rtx_set(ARG0, ARG1) gen_rtx_SET((ARG0), (ARG1))
#endif

#if BUILDING_GCC_VERSION >= 5000 && BUILDING_GCC_VERSION < 10000
template <>
template <>
inline bool is_a_helper<const ggoto *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_GOTO;
}

template <>
template <>
inline bool is_a_helper<const glabel *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_LABEL;
}

template <>
template <>
inline bool is_a_helper<const greturn *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_RETURN;
}

template <>
template <>
inline bool is_a_helper<const gswitch *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_SWITCH;
}
#endif

#ifdef __cplusplus
static inline void debug_tree(const_tree t)
{
	debug_tree(CONST_CAST_TREE(t));
}

static inline void debug_gimple_stmt(const_gimple_stmt_ptr s)
{
	debug_gimple_stmt(CONST_CAST_GIMPLE(s));
}

static inline void debug_bb(const_basic_block bb)
{
	debug_bb(CONST_CAST_BB(bb));
}
#else
#define debug_tree(t) debug_tree(CONST_CAST_TREE(t))
#define debug_gimple_stmt(s) debug_gimple_stmt(CONST_CAST_GIMPLE(s))
#define debug_bb(bb) debug_bb(CONST_CAST_BB(bb))
#endif

#if BUILDING_GCC_VERSION >= 4007 && BUILDING_GCC_VERSION < 7000
static inline source_location expansion_point_location(source_location location)
{
	return linemap_resolve_location(line_table, location, LRK_MACRO_EXPANSION_POINT, NULL);
}
#endif

#if BUILDING_GCC_VERSION < 7000
#define SET_DECL_ALIGN(n, a)	({ DECL_ALIGN(n) = (a); })
#define SET_DECL_MODE(n, m)	({ DECL_MODE(n) = (m); })
#define FOR_EACH_SSA_NAME(I, VAR, FN)	for ((I) = 1; (I) < num_ssa_names; (I)++) if (((VAR) = ssa_name(I)) != NULL_TREE)

static inline unsigned HOST_WIDE_INT least_bit_hwi(unsigned HOST_WIDE_INT x)
{
	return x & -x;
}

static inline bool pow2_or_zerop(unsigned HOST_WIDE_INT x)
{
	return least_bit_hwi(x) == x;
}

static inline bool pow2p_hwi(unsigned HOST_WIDE_INT x)
{
	return x && pow2_or_zerop(x);
}
#endif

#if BUILDING_GCC_VERSION >= 7000
#define get_inner_reference(exp, pbitsize, pbitpos, poffset, pmode, punsignedp, preversep, pvolatilep, keep_aligning)	\
	get_inner_reference(exp, pbitsize, pbitpos, poffset, pmode, punsignedp, preversep, pvolatilep)
#endif

#if BUILDING_GCC_VERSION >= 4007 && BUILDING_GCC_VERSION < 8000
#define ALLOCA_FUNCTION_CODE_P(FCODE)	\
	((FCODE) == BUILT_IN_ALLOCA || (FCODE) == BUILT_IN_ALLOCA_WITH_ALIGN)

static inline void remove_call_arg_locations(rtx_insn *insn)
{
	rtx_insn *next;

	for (next = NEXT_INSN(insn); next; next = NEXT_INSN(next)) {
		rtx_insn *tmp;

		if (GET_CODE(next) == BARRIER)
			continue;
		if (GET_CODE(next) != NOTE)
			break;
		if (NOTE_KIND(next) != NOTE_INSN_CALL_ARG_LOCATION)
			continue;

		tmp = PREV_INSN(next);
		delete_insn(next);
		next = tmp;
	}
}
#endif

#if BUILDING_GCC_VERSION >= 4009 && BUILDING_GCC_VERSION < 8000
#define REG_BR_PROB_NEVER		0
#define REG_BR_PROB_VERY_UNLIKELY	PROB_VERY_UNLIKELY
#define REG_BR_PROB_UNLIKELY		PROB_UNLIKELY
#define REG_BR_PROB_EVEN		PROB_EVEN
#define REG_BR_PROB_LIKELY		PROB_LIKELY
#define REG_BR_PROB_VERY_LIKELY		PROB_VERY_LIKELY
#define REG_BR_PROB_ALWAYS		PROB_ALWAYS
#endif

#if BUILDING_GCC_VERSION >= 5000 && BUILDING_GCC_VERSION < 8000
#define cgraph_create_edge(caller, callee, call_stmt, count, freq, nest) \
	(caller)->create_edge((callee), (call_stmt), (count), (freq))
#define cgraph_create_edge_including_clones(caller, callee, old_call_stmt, call_stmt, count, freq, nest, reason) \
	(caller)->create_edge_including_clones((callee), (old_call_stmt), (call_stmt), (count), (freq), (reason))

typedef unsigned short poly_uint16_pod;
typedef HOST_WIDE_INT poly_int64_pod;
typedef unsigned HOST_WIDE_INT poly_uint64_pod;
typedef offset_int poly_offset_int_pod;
typedef wide_int poly_wide_int_pod;
typedef widest_int poly_widest_int_pod;

typedef unsigned short poly_uint16;
typedef HOST_WIDE_INT poly_int64;
typedef unsigned HOST_WIDE_INT poly_uint64;
typedef offset_int poly_offset_int;
typedef wide_int poly_wide_int;
typedef wide_int_ref poly_wide_int_ref;
typedef widest_int poly_widest_int;
#endif

#if BUILDING_GCC_VERSION < 8000
#define RETURN_FROM_IMM_USE_STMT(ITER, VAL) {	\
	end_imm_use_stmt_traverse(&(ITER));	\
	return (VAL);				\
}

static inline bool multiple_p(poly_int64_pod a, HOST_WIDE_INT b)
{
	return a % b == 0;
}

static inline poly_int64_pod exact_div(poly_int64_pod a, HOST_WIDE_INT b)
{
	return a / b;
}

static inline void print_dec(poly_int64_pod a, FILE *f)
{
	fprintf(f, "%ld", a);
}

#ifdef __cplusplus
static inline void print_gimple_stmt(FILE *file, const_gimple_stmt_ptr g, int spc, int flags)
{
	print_gimple_stmt(file, CONST_CAST_GIMPLE(g), spc, flags);
}

static inline void print_gimple_expr(FILE *file, const_gimple_stmt_ptr g, int spc, int flags)
{
	print_gimple_expr(file, CONST_CAST_GIMPLE(g), spc, flags);
}
#else
#define print_gimple_stmt(file, g, spc, flags) print_gimple_stmt((file), CONST_CAST_GIMPLE(g), (spc), (flags))
#define print_gimple_expr(file, g, spc, flags) print_gimple_expr((file), CONST_CAST_GIMPLE(g), (spc), (flags))
#endif

#define edge_count(e) (e)->count

#define E_BLKmode BLKmode
#define E_BImode BImode
#define E_QImode QImode
#define E_HImode HImode
#define E_SImode SImode
#define E_DImode DImode
#define E_TImode TImode
#define E_OImode OImode
#if BUILDING_GCC_VERSION >= 4009
#define E_XImode XImode
#if BUILDING_GCC_VERSION >= 5000
#define E_BND32mode BND32mode
#define E_BND64mode BND64mode
#endif
#endif
#define E_QQmode QQmode
#define E_HQmode HQmode
#define E_SQmode SQmode
#define E_DQmode DQmode
#define E_TQmode TQmode
#define E_UQQmode UQQmode
#define E_UHQmode UHQmode
#define E_USQmode USQmode
#define E_UDQmode UDQmode
#define E_UTQmode UTQmode
#define E_HAmode HAmode
#define E_SAmode SAmode
#define E_DAmode DAmode
#define E_TAmode TAmode
#define E_UHAmode UHAmode
#define E_USAmode USAmode
#define E_UDAmode UDAmode
#define E_UTAmode UTAmode
#define E_SFmode SFmode
#define E_DFmode DFmode
#define E_XFmode XFmode
#define E_TFmode TFmode
#define E_SDmode SDmode
#define E_DDmode DDmode
#define E_TDmode TDmode

static inline void replace_call_with_value(gimple_stmt_iterator *gsi, tree val)
{
	gimple_stmt_ptr repl, stmt = gsi_stmt(*gsi);
	tree vdef, lhs = gimple_call_lhs(stmt);

	if (lhs) {
		if (!useless_type_conversion_p(TREE_TYPE(lhs), TREE_TYPE(val)))
			val = fold_convert(TREE_TYPE(lhs), val);
		repl = gimple_build_assign(lhs, val);
	} else
		repl = gimple_build_nop();

	vdef = gimple_vdef(stmt);
	if (vdef && TREE_CODE(vdef) == SSA_NAME) {
		unlink_stmt_vdef(stmt);
		release_ssa_name(vdef);
	}

	gsi_replace(gsi, repl, false);
}

static inline bool id_equal(const_tree id, const char *str)
{
	return !strcmp(IDENTIFIER_POINTER(id), str);
}
#endif

#if BUILDING_GCC_VERSION < 8000
#ifdef __cplusplus
#define gen_known_cond_type(cond, op, type)				\
static inline bool known_##cond##_##type(const type& a, const type& b)	\
{									\
	return a op b;							\
}
#else
#define gen_known_cond_type(cond, op, type)				\
static inline bool known_##cond##_##type(type a, type b)		\
{									\
	return a op b;							\
}
#endif
#else
#define gen_known_cond_type(cond, op, type)				\
static inline bool known_##cond##_##type(const type& a, const type& b)	\
{									\
	return known_##cond(a, b);					\
}
#endif

#define gen_known_type(type)			\
	gen_known_cond_type(eq, ==, type)	\
	gen_known_cond_type(ne, !=, type)	\
	gen_known_cond_type(le, <=, type)	\
	gen_known_cond_type(lt, <,  type)	\
	gen_known_cond_type(ge, >=, type)	\
	gen_known_cond_type(gt, >,  type)

gen_known_type(poly_uint16)
gen_known_type(poly_int64)

#undef gen_known_type_cond
#undef gen_known_type

#if BUILDING_GCC_VERSION >= 8000
#define empty_string ""
#define REG_BR_PROB_NEVER		profile_probability::never()
#define REG_BR_PROB_VERY_UNLIKELY	profile_probability::very_unlikely()
#define REG_BR_PROB_UNLIKELY		profile_probability::unlikely()
#define REG_BR_PROB_EVEN		profile_probability::even()
#define REG_BR_PROB_LIKELY		profile_probability::likely()
#define REG_BR_PROB_VERY_LIKELY		profile_probability::very_likely()
#define REG_BR_PROB_ALWAYS		profile_probability::always()
#define edge_count(e) (e)->count()

#define cgraph_create_edge(caller, callee, call_stmt, count, freq, nest) \
	({ (void)(freq); (caller)->create_edge((callee), (call_stmt), (count)); })
#define cgraph_create_edge_including_clones(caller, callee, old_call_stmt, call_stmt, count, freq, nest, reason) \
	({ (void)(freq); ((caller)->create_edge_including_clones((callee), (old_call_stmt), (call_stmt), (count), (reason)); })

#if 0
namespace wi {
	template <>
	struct int_traits <poly_uint16> {
//		static const enum precision_type precision_type = FLEXIBLE_PRECISION;
		static const enum precision_type precision_type = VAR_PRECISION;
		static const bool host_dependent_precision = true;
		static const bool is_sign_extended = true;
		static unsigned int get_precision(const poly_uint16& x) { return 16; }
		static wi::storage_ref decompose(HOST_WIDE_INT *scratch, unsigned int precision, const poly_uint16 &x);
	};
}
#endif

static inline void remove_call_arg_locations(rtx_insn *insn) {}

static inline void print_gimple_stmt(FILE *file, const_gimple_stmt_ptr g, int spc, dump_flags_t flags = TDF_NONE)
{
	print_gimple_stmt(file, CONST_CAST_GIMPLE(g), spc, flags);
}

static inline void print_gimple_expr(FILE *file, const_gimple_stmt_ptr g, int spc, dump_flags_t flags = TDF_NONE)
{
	print_gimple_expr(file, CONST_CAST_GIMPLE(g), spc, flags);
}

// hack to work around a mismatched prototype in gcc headers...
void debug_function(tree fn, dump_flags_t flags);
extern inline void debug_function(tree fn, int flags)
{
	debug_function(fn, (dump_flags_t)flags);
}
#endif

#if BUILDING_GCC_VERSION < 9000
#define NODE_UID(node) (node)->uid
#define fndecl_built_in_p(fndecl) DECL_BUILT_IN(fndecl)
#define fndecl_built_in_class_p(fndecl, klass) (DECL_BUILT_IN(fndecl) && DECL_BUILT_IN_CLASS(fndecl) == (klass))
#define stmt_can_throw_internal(fun, stmt) stmt_can_throw_internal(stmt)
#endif

#if BUILDING_GCC_VERSION >= 9000
#define fndecl_built_in_class_p(fndecl, klass) fndecl_built_in_p((fndecl), (klass))
#define align_functions_log align_functions.levels[0].log
#define NODE_UID(node) (node)->get_uid()

static inline basic_block label_to_block_fn(struct function *ifun, tree dest)
{
	return label_to_block(ifun, dest);
}
#endif

#if BUILDING_GCC_VERSION < 11000
#define NODE_CLONE_INFO(node) (&(node)->clone)
#define NODE_THUNK(node) ((node)->thunk.thunk_p)
#define DECL_IS_UNDECLARED_BUILTIN(decl) DECL_IS_BUILTIN(decl)
#endif

#if BUILDING_GCC_VERSION >= 7000 && BUILDING_GCC_VERSION < 10000
static inline bool ipcp_get_parm_bits(tree parm, tree *value, widest_int *mask)
{
	cgraph_node_ptr cnode;
#if BUILDING_GCC_VERSION >= 9000
	ipcp_transformation *ts;
#else
	ipcp_transformation_summary *ts;
#endif
	tree p;
	int i;
	ipa_bits *bits;

	cnode = cgraph_node::get(current_function_decl);
	ts = ipcp_get_transformation_summary(cnode);
	if (!ts || vec_safe_length(ts->bits) == 0)
		return false;

	p = DECL_ARGUMENTS(current_function_decl);
	i = 0;
	while (p && p != parm) {
		i++;
		p = DECL_CHAIN(p);
	}
	gcc_assert(p);

#if BUILDING_GCC_VERSION >= 10000
	if (NODE_CLONE_INFO(cnode) && NODE_CLONE_INFO(cnode)->param_adjustments) {
		i = NODE_CLONE_INFO(cnode)->param_adjustments->get_original_index(i);
		if (i < 0)
			return false;
	}
#endif

	bits = (*ts->bits)[i];
	if (!bits)
		return false;

	*mask = bits->mask;
	*value = wide_int_to_tree(TREE_TYPE(parm), bits->value);
	return true;
}
#endif

#if BUILDING_GCC_VERSION < 10000
#define NODE_LOCAL(node) (&(node)->local)
#define NODE_GLOBAL(node) (&(node)->global)
#endif

#if BUILDING_GCC_VERSION >= 10000
#define NODE_LOCAL(node) (node)
#define NODE_GLOBAL(node) (node)

rtx maybe_gen_leave(machine_mode);
static inline rtx gen_leave(machine_mode arg0)
{
	rtx res = maybe_gen_leave(arg0);
	gcc_assert(res);
	return res;
}
#endif

#if BUILDING_GCC_VERSION < 11000
#define get_range_info(stmt, expr, min, max) get_range_info((expr), (min), (max))
#endif

#if BUILDING_GCC_VERSION >= 11000
#define VEC(T, A) auto_vec<T>
#define NODE_CLONE_INFO(node) (clone_info::get(node))
#define NODE_THUNK(node) ((node)->thunk)
#define BREAK_FROM_IMM_USE_STMT(iter) break
#define RETURN_FROM_IMM_USE_STMT(iter, val) return (val)

static inline enum value_range_kind get_range_info(gimple_stmt_ptr stmt, tree expr, wide_int *min, wide_int *max);

static inline enum value_range_kind get_range_info(gimple_stmt_ptr stmt, tree name, irange &vr)
{
	tree min, max;
	wide_int wmin, wmax;
	enum value_range_kind kind = get_range_info(stmt, name, &wmin, &wmax);

	if (kind == VR_VARYING)
		vr.set_varying(TREE_TYPE(name));
	else if (kind == VR_UNDEFINED)
		vr.set_undefined();
	else {
		min = wide_int_to_tree(TREE_TYPE(name), wmin);
		max = wide_int_to_tree(TREE_TYPE(name), wmax);
		vr.set(min, max, kind);
	}
	return kind;
}
#endif

#if BUILDING_GCC_VERSION < 12000
static inline bool gimple_deferred_init_p(gimple_stmt_ptr stmt)
{
	return false;
}
#endif

#if BUILDING_GCC_VERSION >= 11000 && BUILDING_GCC_VERSION < 12000
static inline enum value_range_kind get_range_info(gimple_stmt_ptr stmt, tree expr, wide_int *min, wide_int *max)
{
	gimple_ranger rq;
	value_range vr;

	gcc_assert(!POINTER_TYPE_P(TREE_TYPE(expr)));
	gcc_assert(min && max);

	if (TREE_CODE(expr) == INTEGER_CST) {
		*min = wi::to_wide(expr);
		*max = *min;
		return VR_RANGE;
	}
	if (TREE_CODE(expr) != SSA_NAME)
		return VR_VARYING;

	if (!rq.range_of_expr(vr, expr, stmt))
		return VR_VARYING;
	if (vr.undefined_p() || vr.varying_p())
		return VR_VARYING;
	gcc_assert(!vr.symbolic_p());

	*min = wi::to_wide(vr.min());
	*max = wi::to_wide(vr.max());
	return vr.kind();
}
#endif

#if BUILDING_GCC_VERSION >= 12000
#ifdef TARGET_CPU_P
#define TARGET_386
#endif

#define FOR_EACH_LOOP(loop, flags) for (auto loop : loops_list(cfun, (flags)))
#define FOR_EACH_LOOP_FN(fn, loop, flags) for (auto loop : loops_list((fn), (flags)))

static inline bool gimple_deferred_init_p(gimple_stmt_ptr stmt)
{
	if (gimple_assign_single_p(stmt)) {
		tree rhs;

		rhs = gimple_assign_rhs1(stmt);
		if (TREE_CODE(rhs) != SSA_NAME)
			return false;
		stmt = SSA_NAME_DEF_STMT(rhs);
	}

	return gimple_call_internal_p(stmt, IFN_DEFERRED_INIT);
}

static inline enum value_range_kind get_range_info(gimple_stmt_ptr stmt, tree expr, wide_int *min, wide_int *max)
{
	range_query *rq;
	value_range vr;

	gcc_assert(!POINTER_TYPE_P(TREE_TYPE(expr)));
	gcc_assert(min && max);

	if (TREE_CODE(expr) == INTEGER_CST) {
		*min = wi::to_wide(expr);
		*max = *min;
		return VR_RANGE;
	}
	if (TREE_CODE(expr) != SSA_NAME)
		return VR_VARYING;

	rq = get_range_query(cfun);
	if (!rq->range_of_expr(vr, expr, stmt))
		return VR_VARYING;
	if (vr.undefined_p() || vr.varying_p())
		return VR_VARYING;
	gcc_assert(!vr.symbolic_p());

	*min = wi::to_wide(vr.min());
	*max = wi::to_wide(vr.max());
	return vr.kind();
}
#endif

#if BUILDING_GCC_VERSION >= 13000
#ifdef TARGET_386
extern tree ix86_bf16_type_node;
#else
#define ix86_bf16_type_node NULL
#endif
#endif

#if BUILDING_GCC_VERSION < 14000
#define last_nondebug_stmt(bb) last_stmt(bb)
#endif

static inline location_t gimple_expansion_point_location(const_gimple_stmt_ptr g)
{
	return expansion_point_location(gimple_location(g));
}

static inline location_t gimple_inline_point_location(const_gimple_stmt_ptr g)
{
	location_t loc = UNKNOWN_LOCATION;
	tree block;

	block = gimple_block(g);
	gcc_assert(block == NULL_TREE || TREE_CODE(block) == BLOCK);

	for (; block != NULL_TREE && TREE_CODE(block) == BLOCK; block = BLOCK_SUPERCONTEXT(block)) {
		tree ao = BLOCK_ABSTRACT_ORIGIN(block);

		if (ao == NULL_TREE)
			continue;
		if (ao == current_function_decl)
			break;

		switch (TREE_CODE(ao)) {
		case BLOCK:
			if (BLOCK_SOURCE_LOCATION(block) != UNKNOWN_LOCATION)
				loc = BLOCK_SOURCE_LOCATION(block);
			continue;

		case FUNCTION_DECL:
			loc = BLOCK_SOURCE_LOCATION(block);
			continue;

		default:
			debug_gimple_stmt(g);
			debug_tree(block);
			debug_tree(ao);
			gcc_unreachable();
		}
	}
#if BUILDING_GCC_VERSION >= 4006
	gcc_assert(gimple_block(g) == NULL_TREE || block != NULL_TREE);
#endif

	if (loc == UNKNOWN_LOCATION)
		loc = gimple_expansion_point_location(g);
	gcc_assert(loc != UNKNOWN_LOCATION);
	return loc;
}

static inline bool bb_any_loop(basic_block bb)
{
	return bb_loop_depth(bb) || (bb->flags & BB_IRREDUCIBLE_LOOP);
}

static inline tree get_inner_ref(tree arg)
{
	bool done = false;

	while (!done) {
		switch (TREE_CODE(arg)) {
		default:
			done = true;
			break;

#if BUILDING_GCC_VERSION >= 4006
		case MEM_REF:
#else
		case INDIRECT_REF:
#endif
			done = true;
			break;

		case ARRAY_REF:
		case ARRAY_RANGE_REF:
		case BIT_FIELD_REF:
		case COMPONENT_REF:
		case IMAGPART_EXPR:
		case REALPART_EXPR:
		case VIEW_CONVERT_EXPR:
			arg = TREE_OPERAND(arg, 0);
			break;
		}
	}

	return arg;
}

static inline tree get_inner_arg(tree arg)
{
	tree offset;
	poly_int64_pod bitsize, bitpos;
	enum machine_mode mode;
	int unsignedp, reversep, volatilep;

	return get_inner_reference(arg, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
}

static inline bool insn_dominates_insn_p(enum cdi_direction dir, rtx_insn *insn1, rtx_insn *insn2)
{
	basic_block bb1, bb2;

	if (insn1 == insn2)
		return true;

	bb1 = BLOCK_FOR_INSN(insn1);
	bb2 = BLOCK_FOR_INSN(insn2);

	if (bb1 == bb2) {
		rtx_insn *insn;

		FOR_BB_INSNS(bb1, insn) {
			if (insn == insn2)
				return dir == CDI_POST_DOMINATORS;
			if (insn == insn1)
				return dir == CDI_DOMINATORS;
		}

		gcc_unreachable();
	}

	return dominated_by_p(dir, bb2, bb1);
}

static inline struct opt_pass *plugin_find_pass(const char *name)
{
	int i;
#if BUILDING_GCC_VERSION >= 4009
	gcc::pass_manager *passes = g->get_passes();
#endif

#if BUILDING_GCC_VERSION >= 4009
	for (i = 0; i < passes->passes_by_id_size; i++) {
		struct opt_pass *pass = passes->get_pass_for_id(i);
#else
	for (i = 0; i < passes_by_id_size; i++) {
		struct opt_pass *pass = get_pass_for_id(i);
#endif

		if (pass && !strcmp(pass->name, name))
			return pass;
	}

	return NULL;
}

static inline void set_current_function_decl(tree fndecl)
{
	gcc_assert(fndecl != NULL_TREE);

	push_cfun(DECL_STRUCT_FUNCTION(fndecl));
#if BUILDING_GCC_VERSION <= 4007
	current_function_decl = fndecl;
#endif
	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);
}

static inline void unset_current_function_decl(void)
{
	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	pop_cfun();
#if BUILDING_GCC_VERSION <= 4007
	current_function_decl = NULL_TREE;
#endif
}

// create the equivalent of
// asm volatile("" : : : "memory");
// or
// asm("" : "+rm"(var));
// or
// asm("" : "+rm"(var) : : "memory");
static inline gimple_stmt_ptr gimple_barrier(tree var, bool full)
{
	gimple_stmt_ptr stmt;
#if BUILDING_GCC_VERSION <= 4007
	VEC(tree, gc) *inputs = NULL;
	VEC(tree, gc) *outputs = NULL;
	VEC(tree, gc) *clobbers = NULL;
#else
	vec<tree, va_gc> *inputs = NULL;
	vec<tree, va_gc> *outputs = NULL;
	vec<tree, va_gc> *clobbers = NULL;
#endif

	if (var) {
		tree input;
		tree output;

		input = build_tree_list(NULL_TREE, build_const_char_string(2, "0"));
		input = chainon(NULL_TREE, build_tree_list(input, var));
		VEC_safe_push(tree, gc, inputs, input);

		output = build_tree_list(NULL_TREE, build_const_char_string(4, "=rm"));
		if (!VAR_P(var)) {
			gcc_assert(SSA_NAME_VAR(var));
			var = make_ssa_name(SSA_NAME_VAR(var), NULL);
		}
		output = chainon(NULL_TREE, build_tree_list(output, var));
		VEC_safe_push(tree, gc, outputs, output);
	}

	if (full) {
		tree clobber;

		clobber = build_tree_list(NULL_TREE, build_const_char_string(7, "memory"));
		VEC_safe_push(tree, gc, clobbers, clobber);
	}

	stmt = gimple_build_asm_vec("", inputs, outputs, clobbers, NULL);
	if (full && !var)
		gimple_asm_set_volatile(as_a_gasm(stmt), true);
	if (var && !VAR_P(var))
		SSA_NAME_DEF_STMT(var) = stmt;
	return stmt;
}

static inline bool single_arg_phi_p(gimple_stmt_ptr phi_stmt)
{
	if (gimple_code(phi_stmt) != GIMPLE_PHI)
		return false;
	return gimple_phi_num_args(phi_stmt) == 1;
}

static inline bool tree_for_decl_and_aliases(tree decl, bool (*callback)(tree, void *), void *data)
{
	cgraph_node_ptr node;
	unsigned int i;
	alias_pair *p;

	if (callback(decl, data))
		return true;

	if (!alias_pairs)
		return false;

	node = cgraph_get_node(decl);
	if (!node)
		return false;

#if BUILDING_GCC_VERSION <= 4007
	FOR_EACH_VEC_ELT(alias_pair, alias_pairs, i, p) {
#else
	FOR_EACH_VEC_SAFE_ELT(alias_pairs, i, p) {
#endif
		if (TREE_CODE(p->decl) != FUNCTION_DECL)
			continue;
		if (node != cgraph_node_for_asm(p->target))
			continue;

		if (tree_for_decl_and_aliases(p->decl, callback, data))
			return true;
	}

	return false;
}

static inline basic_block get_single_pred_entry_block(void)
{
	basic_block entry_bb;

	entry_bb = ENTRY_BLOCK_PTR_FOR_FN(cfun);
	gcc_assert(single_succ_p(entry_bb));
	entry_bb = single_succ(entry_bb);
	if (single_pred_p(entry_bb))
		return entry_bb;

//	gcc_assert(bb_loop_depth(entry_bb) || (entry_bb->flags & BB_IRREDUCIBLE_LOOP));
	gcc_assert(bb_any_loop(entry_bb));
	entry_bb = plugin_split_edge(single_succ_edge(entry_bb));
	gcc_assert(single_pred_p(entry_bb));
	gcc_assert(single_succ_p(entry_bb));
	gcc_assert(!bb_any_loop(entry_bb));
	return entry_bb;
}

static inline bool has_trait(tree decl, const char *trait)
{
	tree attr;

	attr = lookup_attribute("traits", DECL_ATTRIBUTES(decl));
	if (attr != NULL_TREE)
		attr = TREE_VALUE(attr);

	for (; attr; attr = TREE_CHAIN(attr)) {
		if (strcmp(trait, TREE_STRING_POINTER(TREE_VALUE(attr))))
			continue;

		return true;
	}

	return false;
}

#endif
