/*
 * Kernel constants to make available for plugins without the need to include
 * kernel headers.
 *
 * There's overlap with asm-offsets.h, but we want type preserving symbols, not
 * defines to have proper compiler visibility and be able to use them in
 * plugins.
 *
 * This file's preprocessor output will be post-processed to generate "ksyms.h"
 *
 * Written by Mathias Krause <minipli@grsecurity.net>
 * Copyright 2020 by Open Source Security, Inc.
 * Copyright 2020 by PaX Team
 * Licensed under the GPL v2
 */

#define KSYMS_GENERATE_INCLUDES
#include "ksyms.def"

/* Dummy struct to be able to preserve the original types */
static const struct __plugin_ksyms {
	/* These symbols will be generated as struct fields 'typeof(val) sym;' */
#define DEFINE(sym, val) typeof(val) ksym_##sym;
#define DEFINE_OPTIONAL(sym, val) kifdef_##sym
#define DEFINE_ENUM(name, val) enum ksym_##name ksym_##val;
#define COMMENT(s) COMMENT(s)
#define BLANK() BLANK()
#define KSYMS_GENERATE_DEFINES
#include "ksyms.def"
} __plugin_ksyms __maybe_unused = {
	/* These symbols will be generated as designated initializers '.sym = val,' */
#define DEFINE(sym, val) .ksym_##sym = val,
#define DEFINE_OPTIONAL(sym, val) kdefine_##val
#define DEFINE_ENUM(name, val)
#define COMMENT(s) COMMENT(s)
#define BLANK() BLANK()
#define KSYMS_GENERATE_DEFINES
#include "ksyms.def"
};
