/*
 * Copyright 2022 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * allow specifying traits for plugin gate functions to filter on
 */

#include "utilities.h"

static bool check_traits_attribute(tree decl, const_tree name, const_tree args)
{
	const_tree arg;
	location_t loc;

	loc = DECL_SOURCE_LOCATION(decl);

	for (arg = args; arg; arg = TREE_CHAIN(arg)) {
		tree trait = TREE_VALUE(arg);

		if (TREE_CODE(trait) != STRING_CST) {
			error_at(loc, "attribute %qE requires a string argument instead of %qE", name, trait);
			return false;
		}
	}

	return true;
}

static tree handle_traits_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree decl;

	*no_add_attrs = true;

	decl = *node;
	if (TREE_CODE(decl) != FUNCTION_DECL) {
		error_at(DECL_SOURCE_LOCATION(decl), "attribute %qE applies to functions only", name);
		return NULL_TREE;
	}

	if (!check_traits_attribute(decl, name, args))
		return NULL_TREE;

	*no_add_attrs = false;
	return args;
}

struct attribute_spec traits_attr = ATTRIBUTE_SPEC_INIT("traits", 1, -1, true, false, false, handle_traits_attribute, false, NULL);
